/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-i18n.h>

#include "FileHandler.h"

FileHandler::FileHandler(MainMenu& menu, FileHolder& file_holder, Gtk::Window& parent):
  m_menu(menu), m_file_holder(file_holder), m_parent(parent)
{
  m_menu.action_file_new->signal_activate().connect(sigc::mem_fun(*this, &FileHandler::on_file_new));
  m_menu.action_file_open->signal_activate().connect(sigc::mem_fun(*this, &FileHandler::on_file_open));
  m_menu.action_file_save->signal_activate().connect(sigc::mem_fun(*this, &FileHandler::on_file_save));
  m_menu.action_file_save_as->signal_activate().connect(sigc::mem_fun(*this, &FileHandler::on_file_save_as));
  m_menu.action_file_save_all->signal_activate().connect(sigc::mem_fun(*this, &FileHandler::on_file_save_all));
  m_menu.action_file_close->signal_activate().connect(sigc::mem_fun(*this, &FileHandler::on_file_close));

  file_holder.signal_document_changed().connect(sigc::mem_fun(*this, &FileHandler::on_document_changed));
  file_holder.signal_document_close().connect(sigc::mem_fun(*this, &FileHandler::on_document_close));
}

bool FileHandler::new_document()
{
  m_file_holder.create_document();
  return true; // Always success
}

bool FileHandler::open_document()
{
  const std::string path = select_open_path();
  if(path.empty() ) return false;

  return open_document_from(path);
}

bool FileHandler::open_document_from(const std::string& path)
{
  try
  {
    m_file_holder.create_document(path);
  }
  catch(Glib::FileError& e)
  {
    Gtk::MessageDialog dialog(m_parent, e.what(), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    dialog.set_title(_("Failed to open file"));
    dialog.run();
    return false;
  }
  catch(Glib::ConvertError& e)
  {
    Gtk::MessageDialog dialog(m_parent, _("Could not detect character encoding of input file. Either the encoding is unknown, or it is a binary file."), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    dialog.set_title(_("Failed to open file"));
    dialog.run();
    return false;
  }

  return true;
}

bool FileHandler::save_document(C4sDocument* document)
{
  if(c4s_document_get_path(document) != NULL)
  {
    GError* error = NULL;
    if(c4s_document_save_file(document, c4s_document_get_path(document), &error) == FALSE)
    {
      Gtk::MessageDialog dialog(m_parent, error->message, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
      dialog.run();
      g_error_free(error);
      return false;
    }
  }
  else
  {
    return save_document_as(document);
  }

  return true;
}

bool FileHandler::save_document_as(C4sDocument* document)
{
  const std::string path = select_save_path(document);
  if(path.empty() ) return false;

  GError* error = NULL;
  if(c4s_document_save_file(document, path.c_str(), &error) == FALSE)
  {
    Gtk::MessageDialog dialog(m_parent, error->message, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    g_error_free(error);
    dialog.run();
    return false;
  }

  return true;
}

bool FileHandler::save_all_documents()
{
  bool result = true;
  for(int i = 0; i < m_file_holder.get_n_pages(); ++ i)
  {
    if(!save_document(c4s_document_window_get_document(C4S_DOCUMENT_WINDOW(m_file_holder.get_nth_page(i)->gobj()))))
      result = false;
  }

  return result;
}

bool FileHandler::close_document(C4sDocument* document)
{
  if(m_close_confirmation.get() == NULL)
    m_close_confirmation.reset(new CloseConfirmationDialog(m_parent));

  m_close_confirmation->add_document(document);
  return launch_close_confirmation();
}

bool FileHandler::close_all_documents()
{
  if(m_close_confirmation.get() == NULL)
    m_close_confirmation.reset(new CloseConfirmationDialog(m_parent));

  for(int i = 0; i < m_file_holder.get_n_pages(); ++ i)
  {
    C4sDocument* doc = c4s_document_window_get_document(C4S_DOCUMENT_WINDOW(m_file_holder.get_nth_page(i)->gobj()));
    m_close_confirmation->add_document(doc);
  }

  return launch_close_confirmation();
}

void FileHandler::on_document_changed(C4sDocumentWindow* window)
{
  m_menu.action_file_save->set_sensitive(window != NULL);
  m_menu.action_file_save_as->set_sensitive(window != NULL);
  m_menu.action_file_save_all->set_sensitive(window != NULL);
  m_menu.action_file_close->set_sensitive(window != NULL);
}

void FileHandler::on_document_close(C4sDocumentWindow* window)
{
  close_document(c4s_document_window_get_document(window));
}

void FileHandler::on_file_new()
{
  m_file_holder.create_document();
}

void FileHandler::on_file_open()
{
  open_document();
}

void FileHandler::on_file_save()
{
  C4sDocument* doc = c4s_document_window_get_document(m_file_holder.get_current_document());
  save_document(doc);
}

void FileHandler::on_file_save_as()
{
  C4sDocument* doc = c4s_document_window_get_document(m_file_holder.get_current_document());
  save_document_as(doc);
}

void FileHandler::on_file_save_all()
{
  save_all_documents();
}

void FileHandler::on_file_close()
{
  C4sDocument* doc = c4s_document_window_get_document(m_file_holder.get_current_document());
  close_document(doc);
}

std::string FileHandler::select_save_path(C4sDocument* document)
{
  Gtk::FileChooserDialog dialog(m_parent, _("Choose a location where to store the script to"), Gtk::FILE_CHOOSER_ACTION_SAVE);
  dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  dialog.add_button(Gtk::Stock::SAVE, Gtk::RESPONSE_APPLY);

  if(c4s_document_get_path(document) != NULL)
  {
    dialog.set_filename(c4s_document_get_path(document));
  }
  else
  {
    dialog.set_current_name("Script.c");
    if(!m_last_path.empty() )
      dialog.set_current_folder(m_last_path);
  }

  dialog.set_modal(true);
  dialog.set_do_overwrite_confirmation(true);
  if(dialog.run() == Gtk::RESPONSE_APPLY)
  {
    m_last_path = dialog.get_current_folder();
    return dialog.get_filename();
  }

  return "";
}

std::string FileHandler::select_open_path()
{
  Gtk::FileChooserDialog dialog(m_parent, _("Choose a script to open"), Gtk::FILE_CHOOSER_ACTION_OPEN);
  dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_APPLY);

  if(!m_last_path.empty() )
    dialog.set_current_folder(m_last_path);

  dialog.set_modal(true);
  if(dialog.run() == Gtk::RESPONSE_APPLY)
  {
    m_last_path = dialog.get_current_folder();
    return dialog.get_filename();
  }

  return "";
}

bool FileHandler::launch_close_confirmation()
{
  CloseConfirmationDialog::Action action = m_close_confirmation->launch();
  if(action == CloseConfirmationDialog::CANCEL)
  {
    m_close_confirmation->clear();
    return false;
  }

  bool result = true;
  std::list<C4sDocumentWindow*> documents_to_close;

  for(CloseConfirmationDialog::iterator iter = m_close_confirmation->begin(); iter != m_close_confirmation->end(); ++ iter)
  {
    if(action == CloseConfirmationDialog::SAVE_SELECTED)
    {
      if(iter.get_save())
      {
        if(!save_document(iter.get_document()))
        {
          // Stop whole process after first error
	  result = false;
	  break;
        }
      }
    }

    // TODO: Need to get DocumentWindow... quite ugly here.
    // CloseConfirmationDialog should save DocumentWindows.
    for(int i = 0; i < m_file_holder.get_n_pages(); ++ i)
    {
      C4sDocumentWindow* window = C4S_DOCUMENT_WINDOW(m_file_holder.get_nth_page(i)->gobj());
      if(c4s_document_window_get_document(window) == iter.get_document())
      {
        documents_to_close.push_back(window);
        break;
      }
    }
  }

  // Prepare for reuse
  m_close_confirmation->clear();

  // Delete required documents
  for(std::list<C4sDocumentWindow*>::iterator iter = documents_to_close.begin(); iter != documents_to_close.end(); ++ iter)
  {
    m_file_holder.remove_document(*iter);
  }

  return result;
}
