/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-documentwindow.h>
#include <c4scripter/c4s-i18n.h>

#include "FindDialog.h"
#include <gtksourceview/gtksourceiter.h>

#include <string.h>

namespace
{
  const int RESPONSE_FIND = 10;
  const int RESPONSE_REPLACE = 11;
  const int RESPONSE_REPLACE_ALL = 12;
}

FindDialog::FindDialog(Gtk::Window& parent, FileHolder& file_holder):
  Gtk::Dialog("",parent),
  m_file_holder(file_holder),
  m_find_label(_("Search for:"), Gtk::ALIGN_RIGHT),
  m_replace_label(_("Replace with:"), Gtk::ALIGN_RIGHT),
  m_case_sensitive(_("Match case")),
  m_whole_word(_("Match entire word only")),
  m_backwards(_("Search backwards")),
  m_wrap_around(_("Wrap around"))
{
  set_type_hint(Gdk::WINDOW_TYPE_HINT_DIALOG);
  set_transient_for(parent);
  set_position(Gtk::WIN_POS_CENTER_ON_PARENT);

  m_find_entry.set_activates_default(true);
  m_replace_entry.set_activates_default(true);

  m_search_table.attach(m_find_label, 0, 1, 0, 1, Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK);
  m_search_table.attach(m_replace_label, 0, 1, 1, 2, Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK);
  m_search_table.attach(m_find_entry, 1, 2, 0, 1, Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK);
  m_search_table.attach(m_replace_entry, 1, 2, 1, 2, Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK);
  m_search_table.set_spacings(10);

  get_vbox()->pack_start(m_search_table, Gtk::PACK_SHRINK);
  get_vbox()->pack_start(m_case_sensitive, Gtk::PACK_SHRINK);
  get_vbox()->pack_start(m_whole_word, Gtk::PACK_SHRINK);
  get_vbox()->pack_start(m_backwards, Gtk::PACK_SHRINK);
  get_vbox()->pack_start(m_wrap_around, Gtk::PACK_SHRINK);
  get_vbox()->set_spacing(10);

  set_border_width(10);

  m_wrap_around.set_active(true);

  add_button(Gtk::Stock::CLOSE, Gtk::RESPONSE_CLOSE);
  m_replace_all_button = add_button(_("Replace _All"), RESPONSE_REPLACE_ALL);
  m_replace_button = add_button(_("_Replace"), RESPONSE_REPLACE);
  add_button(Gtk::Stock::FIND, RESPONSE_FIND);

  m_replace_button->set_image(*Gtk::manage(new Gtk::Image(Gtk::Stock::FIND_AND_REPLACE, Gtk::ICON_SIZE_BUTTON)));
  m_replace_all_button->set_image(*Gtk::manage(new Gtk::Image(Gtk::Stock::FIND_AND_REPLACE, Gtk::ICON_SIZE_BUTTON)));

  set_resizable(false);
  show_all_children();
}

void FindDialog::set_search_only(bool search_only)
{
  if(search_only)
  {
    set_default_response(RESPONSE_FIND);

    m_replace_button->hide();
    m_replace_all_button->hide();
    m_replace_label.hide();
    m_replace_entry.hide();

    set_title(_("Search"));
  }
  else
  {
    set_default_response(RESPONSE_REPLACE);

    m_replace_button->show();
    m_replace_all_button->show();
    m_replace_label.show();
    m_replace_entry.show();

    set_title(_("Search and replace"));
  }
}

void FindDialog::on_response(int response_id)
{
  switch(response_id)
  {
  case Gtk::RESPONSE_CLOSE:
    hide();
    break;
  case RESPONSE_FIND:
    find();
    break;
  case RESPONSE_REPLACE:
    replace();
    break;
  case RESPONSE_REPLACE_ALL:
    replace_all();
    break;
  default:
    Gtk::Dialog::on_response(response_id);
    break;
  }
}

void FindDialog::on_show()
{
  Gtk::Dialog::on_show();
  m_find_entry.grab_focus();
}

bool FindDialog::search_once(const Gtk::TextIter& from, const Gtk::TextIter* to, Gtk::TextIter& match_start, Gtk::TextIter& match_end)
{
  GtkSourceSearchFlags flags = GtkSourceSearchFlags(0);
  if(!m_case_sensitive.get_active() )
    flags = GTK_SOURCE_SEARCH_CASE_INSENSITIVE;

  gboolean (*search_func)(const GtkTextIter*, const gchar*, GtkSourceSearchFlags, GtkTextIter*, GtkTextIter*, const GtkTextIter*) = gtk_source_iter_forward_search;

  if(m_backwards.get_active() )
    search_func = gtk_source_iter_backward_search;

  Glib::ustring find_str = m_find_entry.get_text();
  GtkTextIter match_start_gtk, match_end_gtk;

  if(search_func(from.gobj(), find_str.c_str(), flags, &match_start_gtk, &match_end_gtk, to != NULL ? to->gobj() : NULL) == TRUE)
  {
    match_start = Gtk::TextIter(&match_start_gtk);
    match_end = Gtk::TextIter(&match_end_gtk);

    return true;
  }

  return false;
}

bool FindDialog::search_range(const Gtk::TextIter& from, const Gtk::TextIter* to, Gtk::TextIter& match_start, Gtk::TextIter& match_end)
{
	Gtk::TextIter start_pos(from);
	while(search_once(start_pos, to, match_start, match_end))
	{
	  // search_once does not check for whole word
	  if(m_whole_word.get_active() )
	  {
	    if(!match_start.starts_word() || !match_end.ends_word())
	    {
	      if(m_backwards.get_active() )
	        start_pos = match_start;
	      else
	        start_pos = match_end;

	      continue;
	    }
	  }

	  return true;
	}

	return false;
}

bool FindDialog::search_wrap(const Gtk::TextIter& from, Gtk::TextIter& match_start, Gtk::TextIter& match_end)
{
	Glib::RefPtr<Gtk::TextBuffer> buffer = from.get_buffer();
	Gtk::TextIter start_pos(from);
	
	if(search_range(start_pos, NULL, match_start, match_end))
	  return true;

  if(!m_wrap_around.get_active()) return false;

  Gtk::TextIter restart_pos;
  if(m_backwards.get_active() )
    restart_pos = buffer->end();
  else
    restart_pos = buffer->begin();

  Gtk::TextIter* relimit = &start_pos;
  if(!m_backwards.get_active() )
  {
    start_pos.forward_chars(m_find_entry.get_text().length() );
    if(start_pos == buffer->end() )
      relimit = NULL;
  }
  
  return search_range(restart_pos, relimit, match_start, match_end);
}

bool FindDialog::search_sel(const Gtk::TextIter& from)
{
  C4sDocumentWindow* window = m_file_holder.get_current_document();
  if(window == NULL) return false;

  // Searches from sel and selects the result, if any
  Gtk::TextView* view = Glib::wrap(GTK_TEXT_VIEW(c4s_document_window_get_script_view(window)));
  Glib::RefPtr<Gtk::TextBuffer> buffer = Glib::wrap(GTK_TEXT_BUFFER(c4s_document_window_get_document(window)), TRUE);

  Gtk::TextIter match_start, match_end;
  if(search_wrap(from, match_start, match_end))
  {
    if(m_backwards.get_active() )
      buffer->select_range(match_start, match_end);
    else
      buffer->select_range(match_end, match_start);

    view->scroll_to(buffer->get_insert(), 0.1);
    return true;
  }

  return false;
}

void FindDialog::find()
{
  C4sDocumentWindow* window = m_file_holder.get_current_document();
  if(window == NULL) return;

  Glib::RefPtr<Gtk::TextBuffer> buffer = Glib::wrap(GTK_TEXT_BUFFER(c4s_document_window_get_document(window)), TRUE);
  if(!search_sel(buffer->get_insert()->get_iter()))
  {
    gchar* text;
    std::string str = m_find_entry.get_text();
    text = g_strdup_printf(_("The text \"%s\" was not found."), str.c_str());
    Gtk::MessageDialog dialog(*this, text, false, Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, true);
    dialog.set_title(_("No occurences found"));
    dialog.run();
    g_free(text);
  }
}

void FindDialog::replace()
{
  C4sDocumentWindow* window = m_file_holder.get_current_document();
  if(window == NULL) return;

  Glib::RefPtr<Gtk::TextBuffer> buffer = Glib::wrap(GTK_TEXT_BUFFER(c4s_document_window_get_document(window)), TRUE);

  Gtk::TextIter sel_begin, sel_end;
  buffer->get_selection_bounds(sel_begin, sel_end);

  std::string selected_text = sel_begin.get_slice(sel_end);
  std::string find_text = m_find_entry.get_text();

  bool match;

  if(!m_case_sensitive.get_active() )
  {
    gchar* sel_text_casefold = g_utf8_casefold(selected_text.c_str(), -1);
    gchar* find_text_casefold = g_utf8_casefold(find_text.c_str(), -1);

    match = (strcmp(sel_text_casefold, find_text_casefold) == 0);

    g_free(sel_text_casefold);
    g_free(find_text_casefold);
  }
  else
  {
    match = (selected_text == find_text);
  }
  
  if(match)
  {
    buffer->erase_selection();
    buffer->insert_at_cursor(m_replace_entry.get_text() );

    // Find next, but do not use on_find for this purpose because we do not want
    // to show an error message since we already replaced something. on_find will
    // be used next time when the user retries to press replace.
    search_sel(buffer->get_insert()->get_iter());
  }
  else
  {
    // No match, find first one
    find();
  }

#if 0
  Glib::RefPtr<Gtk::TextBuffer> buffer = m_file_holder.get_current_document()->get_document().get_text_buffer();
  /*
  //bool 	get_has_selection () const, hmm
  if(!buffer->get_has_selection()) //STIRP DOOFE BEDINGUNG.
    return;
  */
  Gtk::TextIter start;
  start = buffer->get_insert()->get_iter();
  buffer->erase_selection();
  buffer->insert_at_cursor(m_replace_entry.get_text());
  buffer->select_range(start, buffer->get_insert()->get_iter());
#endif
}

void FindDialog::replace_all()
{
  C4sDocumentWindow* window = m_file_holder.get_current_document();
  if(window == NULL) return;

  Glib::RefPtr<Gtk::TextBuffer> buffer = Glib::wrap(GTK_TEXT_BUFFER(c4s_document_window_get_document(window)), TRUE);

  Gtk::TextIter begin = buffer->begin();
  unsigned int replace_count = 0;
  Gtk::TextIter match_start, match_end;
  
  while(search_range(begin, NULL, match_start, match_end))
  {
    begin = buffer->erase(match_start, match_end);
    begin = buffer->insert(begin, m_replace_entry.get_text());

    ++ replace_count;
  }

  std::string message;

  gchar* text;
  switch(replace_count)
  {
  case 0:
    message = _("No occurence has been replaced");
    break;
  case 1:
    message = _("1 occurence has been replaced");
    break;
  default:
    text = g_strdup_printf(_("%u occurences have been replaced"), replace_count);
    message = text;
    g_free(text);
    break;
  }

  Gtk::MessageDialog dialog(*this, message, false, Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, true);
  dialog.set_title(_("Replacement result"));
  dialog.run();
}
