/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-i18n.h>

#include "Statusbar.h"

#include <string.h>

Statusbar::Statusbar(FileHolder& file_holder, C4sPreferences* preferences):
  m_file_holder(file_holder),
  m_preferences(preferences),
  m_current_document(NULL)
{
  m_cursor_position_bar.set_size_request(160, 10);
  m_cursor_position_bar.set_has_resize_grip(true);

  set_has_resize_grip(false);
  pack_end(m_cursor_position_bar, Gtk::PACK_SHRINK, 0);

  g_signal_connect(
    G_OBJECT(m_preferences),
    "notify",
    G_CALLBACK(on_notify),
    this
  );

  m_file_holder.signal_document_changed().connect(sigc::mem_fun(*this, &Statusbar::on_document_changed));

  on_document_changed(m_file_holder.get_current_document());
  on_statusbar_visibility();
}

void Statusbar::on_notify(GObject* object, GParamSpec* pspec, gpointer data)
{
  if(strcmp(pspec->name, "view-statusbar") == 0)
    static_cast<Statusbar*>(data)->on_statusbar_visibility();
}

void Statusbar::on_statusbar_visibility()
{
  if(c4s_preferences_get_view_statusbar(m_preferences) == TRUE)
    show();
  else
    hide();
}

void Statusbar::on_document_changed(C4sDocumentWindow* window)
{
  if(m_conn_cursor_moved.connected()) m_conn_cursor_moved.disconnect();
  if(m_conn_text_inserted.connected()) m_conn_text_inserted.disconnect();
  if(m_conn_text_erased.connected()) m_conn_text_erased.disconnect();

  if(window == NULL)
  {
    m_current_document = NULL;
    update_statusbar(NULL);
  }
  else
  {
    m_current_document = c4s_document_window_get_document(window);

    g_signal_connect_after(
      G_OBJECT(m_current_document),
      "mark-set",
      G_CALLBACK(on_cursor_moved_static),
      this
    );

    g_signal_connect_after(
      G_OBJECT(m_current_document),
      "insert-text",
      G_CALLBACK(on_text_inserted_static),
      this
    );

    g_signal_connect_after(
      G_OBJECT(m_current_document),
      "delete-range",
      G_CALLBACK(on_text_erased_static),
      this
    );

    GtkTextIter insert;
    gtk_text_buffer_get_iter_at_mark(GTK_TEXT_BUFFER(m_current_document), &insert, gtk_text_buffer_get_insert(GTK_TEXT_BUFFER(m_current_document)));
    Gtk::TextIter cpp(&insert);
    update_statusbar(&cpp);
  }
}

void Statusbar::on_cursor_moved(GtkTextIter* iter, GtkTextMark* mark)
{
  if(mark == gtk_text_buffer_get_insert(GTK_TEXT_BUFFER(m_current_document)))
  {
    Gtk::TextIter cpp(iter);
    update_statusbar(&cpp);
  }
}

void Statusbar::on_text_inserted(GtkTextIter* iter, const gchar* text, int bytes)
{
  Gtk::TextIter cpp(iter);
  update_statusbar(&cpp);
}

void Statusbar::on_text_erased(GtkTextIter* begin, GtkTextIter* end)
{
  // begin == end
  Gtk::TextIter cpp(begin);
  update_statusbar(&cpp);
}

void Statusbar::update_statusbar(const Gtk::TextIter* pos)
{
  m_cursor_position_bar.pop();
  if(pos == NULL) return;

  gchar* text = g_strdup_printf(_("Line %d, Column %d"), pos->get_line() + 1, pos->get_line_offset() + 1);
  m_cursor_position_bar.push(text);
  g_free(text);
}
