/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-document.h>
#include <c4scripter/c4s-highlighter.h>
#include <c4scripter/c4s-textutil.h>
#include <c4scripter/c4s-marshal.h>
#include <c4scripter/c4s-i18n.h>

#include <string.h>

typedef struct _C4sDocumentPrivate C4sDocumentPrivate;
struct _C4sDocumentPrivate {
  C4sPreferences* preferences;

  gchar* title;
  gchar* path;

  C4sDocumentLineEndings line_endings;
  const gchar* encoding;
};

#define C4S_DOCUMENT_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_DOCUMENT, \
    C4sDocumentPrivate \
  ))

enum {
  PROP_0,

  PROP_PREFERENCES,

  PROP_TITLE,
  PROP_PATH,

  PROP_LINE_ENDINGS,
  PROP_ENCODING
};

enum {
  TITLE_CHANGED,
  PATH_CHANGED,
  
  LAST_SIGNAL
};

static const gchar* C4S_DOCUMENT_ENCODING_LATIN1 = "LATIN1";
static const gchar* C4S_DOCUMENT_ENCODING_UTF8 = "UTF-8";

static GtkSourceBufferClass* parent_class = NULL;
static guint document_signals[LAST_SIGNAL];

static gchar*
c4s_document_get_title_from_path(const gchar* path)
{
  gchar* title;
  gchar* converted;

  gchar* dir;
  gchar* dir_converted;
  gchar* sep;

  title = g_path_get_basename(path);
  converted = g_filename_display_name(title);
  g_free(title);

  if(g_ascii_strcasecmp(converted, "Script.c") == 0)
  {
    /* Most clonk scripts are named "Script.c", but we do not want to have
     * thousands of different files titled Script.c. */
    title = g_path_get_dirname(path);
    dir = g_path_get_basename(title);
    g_free(title);
    dir_converted = g_filename_display_name(dir);
    g_free(dir);

    /* If the script is in a folder ending on ".c4d" or ".c4s", we use the
     * object (or scenario, respectively) filename as title. */
    sep = strrchr(dir_converted, '.');
    if(sep != NULL)
    {
      ++ sep;
      if(g_ascii_strcasecmp(sep, "c4s") == 0 ||
         g_ascii_strcasecmp(sep, "c4d") == 0)
      {
        g_free(converted);
        return dir_converted;
      }
    }

    g_free(dir_converted);
    return converted;
  }
  else
  {
    return converted;
  }
}

static gboolean
c4s_document_write_stream(GIOChannel* channel,
                          const gchar* text,
                          gsize length,
                          GError** error)
{
  GIOStatus status;
  gsize written;
  gsize current;

  written = 0;

  while(written < length)
  {
    status = g_io_channel_write_chars(
      channel,
      text + written,
      length - written,
      &current,
      error
    );

    switch(status == G_IO_STATUS_NORMAL)
    {
    case G_IO_STATUS_NORMAL:
      written += current;
      break;
    case G_IO_STATUS_ERROR:
      return FALSE;
    }
  }

  return TRUE;
}

static gboolean
c4s_document_load_stream(C4sDocument* document,
                         GIOChannel* channel,
                         const gchar* encoding,
                         gboolean* is_ascii,
                         GError** error)
{
  C4sDocumentPrivate* priv;
  GIOStatus status;
  gchar* line;
  gsize length;
  gsize terminator;
  GtkTextIter iter;
  gboolean ascii;

  priv = C4S_DOCUMENT_PRIVATE(document);

  do
  {
    status = g_io_channel_seek_position(channel, 0, G_SEEK_SET, error);
    if(status == G_IO_STATUS_ERROR) return FALSE;
  } while(status == G_IO_STATUS_AGAIN);

  do
  {
    status = g_io_channel_set_encoding(channel, encoding, error);
    if(status == G_IO_STATUS_ERROR) return FALSE;
  } while(status == G_IO_STATUS_AGAIN);

  gtk_text_buffer_set_text(GTK_TEXT_BUFFER(document), "", 0);
  gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(document), &iter);

  ascii = TRUE;

  do
  {
    status = g_io_channel_read_line(
      channel,
      &line,
      &length,
      &terminator,
      error
    );

    if(status == G_IO_STATUS_NORMAL)
    {
      if(length > terminator)
      {
        switch(line[terminator])
        {
        case '\n':
          priv->line_endings = C4S_DOCUMENT_LINE_ENDINGS_UNIX;
          break;
        case '\r':
          if(length > terminator + 1 && line[terminator + 1] == '\n')
            priv->line_endings = C4S_DOCUMENT_LINE_ENDINGS_DOS;
          else
            priv->line_endings = C4S_DOCUMENT_LINE_ENDINGS_MAC;
          break;
        default:
          break;
        }

        /* Always terminate with '\n' for loading into GtkTextBuffer */
        line[terminator] = '\n';
        line[terminator + 1] = '\0';
        length = terminator + 1;
      }

      if(ascii == TRUE)
        ascii = c4s_text_is_ascii(line);

      gtk_text_buffer_insert(
        GTK_TEXT_BUFFER(document),
        &iter,
        line,
        length
      );
    }
  } while(status != G_IO_STATUS_EOF && status != G_IO_STATUS_ERROR);

  if(status == G_IO_STATUS_ERROR) return FALSE;

  if(is_ascii != NULL) *is_ascii = ascii;
  g_object_notify(G_OBJECT(document), "line-endings");
  return TRUE;
}

static gboolean
c4s_document_save_stream(C4sDocument* document,
                         GIOChannel* channel,
                         const gchar* encoding,
                         GError** error)
{
  C4sDocumentPrivate* priv;
  GIOStatus status;
  gchar* text;
  gsize length;
  GtkTextIter prev;
  GtkTextIter pos;
  gboolean result;

  priv = C4S_DOCUMENT_PRIVATE(document);

  do
  {
    status = g_io_channel_seek_position(channel, 0, G_SEEK_SET, error);
    if(status == G_IO_STATUS_ERROR) return FALSE;
  } while(status == G_IO_STATUS_AGAIN);

  do
  {
    status = g_io_channel_set_encoding(channel, encoding, error);
    if(status == G_IO_STATUS_ERROR) return FALSE;
  } while(status == G_IO_STATUS_AGAIN);

  gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(document), &prev);
  pos = prev;

  do
  {
    if(!gtk_text_iter_ends_line(&pos))
      gtk_text_iter_forward_to_line_end(&pos);

    text = gtk_text_buffer_get_text(
      GTK_TEXT_BUFFER(document),
      &prev,
      &pos,
      FALSE
    );

    gtk_text_iter_forward_line(&pos);

    length = strlen(text);

    result = c4s_document_write_stream(channel, text, length, error);
    g_free(text);

    if(result == TRUE)
    {
      switch(priv->line_endings)
      {
      case C4S_DOCUMENT_LINE_ENDINGS_DOS:
        result = c4s_document_write_stream(channel, "\r\n", 2, error);
        break;
      case C4S_DOCUMENT_LINE_ENDINGS_UNIX:
        result = c4s_document_write_stream(channel, "\n", 1, error);
        break;
      case C4S_DOCUMENT_LINE_ENDINGS_MAC:
        result = c4s_document_write_stream(channel, "\r", 1, error);
        break;
      default:
        g_assert_not_reached();
        break;
      }
    }
  } while(result == TRUE && gtk_text_iter_forward_line(&prev) == TRUE);

  return result;
}

static void
c4s_document_notify_highlight_brackets_cb(C4sPreferences* preferences,
                                          GParamSpec* pspec,
                                          gpointer user_data)
{
  gtk_source_buffer_set_highlight_matching_brackets(
    GTK_SOURCE_BUFFER(user_data),
    c4s_preferences_get_highlight_brackets(preferences)
  );
}

static void
c4s_document_set_preferences(C4sDocument* document,
                             C4sPreferences* preferences)
{
  C4sDocumentPrivate* priv;
  priv = C4S_DOCUMENT_PRIVATE(document);

  if(priv->preferences != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_document_notify_highlight_brackets_cb),
      document
    );

    g_object_unref(G_OBJECT(priv->preferences));
  }

  priv->preferences = preferences;

  if(preferences != NULL)
  {
    g_object_ref(G_OBJECT(preferences));

    g_signal_connect(
      G_OBJECT(preferences),
      "notify::highlight-brackets",
      G_CALLBACK(c4s_document_notify_highlight_brackets_cb),
      document
    );

    gtk_source_buffer_set_highlight_matching_brackets(
      GTK_SOURCE_BUFFER(document),
      c4s_preferences_get_highlight_brackets(preferences)
    );
  }
}

static void
c4s_document_init(C4sDocument* bar)
{
  C4sDocumentPrivate* priv;
  priv = C4S_DOCUMENT_PRIVATE(bar);

  priv->preferences = NULL;

  priv->title = g_strdup("Untitled");
  priv->path = NULL;

  priv->line_endings = C4S_DOCUMENT_LINE_ENDINGS_DOS;
  priv->encoding = C4S_DOCUMENT_ENCODING_LATIN1;
}

static void
c4s_document_dispose(GObject* object)
{
  C4sDocument* document;
  C4sDocumentPrivate* priv;

  document = C4S_DOCUMENT(object);
  priv = C4S_DOCUMENT_PRIVATE(document);

  c4s_document_set_preferences(document, NULL);

  if(G_OBJECT_CLASS(parent_class)->dispose)
    G_OBJECT_CLASS(parent_class)->dispose(object);
}

static void 
c4s_document_finalize(GObject* object)
{
  C4sDocument* document;
  C4sDocumentPrivate* priv;

  document = C4S_DOCUMENT(object);
  priv = C4S_DOCUMENT_PRIVATE(document);

  g_free(priv->title);
  g_free(priv->path);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_document_set_property(GObject* object,
                          guint prop_id,
                          const GValue* value,
                          GParamSpec* pspec)
{
  C4sDocument* document;
  C4sDocumentPrivate* priv;
  const gchar* encoding;

  g_return_if_fail(C4S_IS_DOCUMENT(object));

  document = C4S_DOCUMENT(object);
  priv = C4S_DOCUMENT_PRIVATE(document);

  switch(prop_id)
  {
  case PROP_PREFERENCES:
    c4s_document_set_preferences(
      document,
      C4S_PREFERENCES(g_value_get_object(value))
    );

    break;
  case PROP_TITLE:
    g_signal_emit(
      G_OBJECT(document),
      document_signals[TITLE_CHANGED],
      0,
      g_value_get_string(value)
    );

    break;
  case PROP_PATH:
    g_signal_emit(
      G_OBJECT(document),
      document_signals[PATH_CHANGED],
      0,
      g_value_get_string(value)
    );

    break;

  case PROP_LINE_ENDINGS:
    priv->line_endings = g_value_get_enum(value);
    break;
  case PROP_ENCODING:
    encoding = g_value_get_string(value);
    if(strcmp(encoding, C4S_DOCUMENT_ENCODING_LATIN1) == 0)
      priv->encoding = C4S_DOCUMENT_ENCODING_LATIN1;
    else if(strcmp(encoding, C4S_DOCUMENT_ENCODING_UTF8) == 0)
      priv->encoding = C4S_DOCUMENT_ENCODING_UTF8;

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_document_get_property(GObject* object,
                          guint prop_id,
                          GValue* value,
                          GParamSpec* pspec)
{
  C4sDocument* document;
  C4sDocumentPrivate* priv;

  g_return_if_fail(C4S_IS_DOCUMENT(object));

  document = C4S_DOCUMENT(object);
  priv = C4S_DOCUMENT_PRIVATE(document);

  switch(prop_id)
  {
  case PROP_PREFERENCES:
    g_value_set_object(value, G_OBJECT(priv->preferences));
    break;
  case PROP_TITLE:
    g_value_set_string(value, priv->title);
    break;
  case PROP_PATH:
    g_value_set_string(value, priv->path);
    break;
  case PROP_LINE_ENDINGS:
    g_value_set_enum(value, priv->line_endings);
    break;
  case PROP_ENCODING:
    g_value_set_static_string(value, priv->encoding);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_document_title_changed(C4sDocument* document,
                           const gchar* title)
{
  C4sDocumentPrivate* priv;
  priv = C4S_DOCUMENT_PRIVATE(document);

  g_free(priv->title);
  priv->title = g_strdup(title);

  g_object_notify(G_OBJECT(document), "title");
}

static void
c4s_document_path_changed(C4sDocument* document,
                          const gchar* path)
{
  C4sDocumentPrivate* priv;
  priv = C4S_DOCUMENT_PRIVATE(document);

  g_free(priv->path);
  if(path != NULL)
    priv->path = g_strdup(path);
  else
    priv->path = NULL;

  g_object_notify(G_OBJECT(document), "path");
}

static void
c4s_document_class_init(C4sDocumentClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = GTK_SOURCE_BUFFER_CLASS(g_type_class_peek_parent(klass));

  g_type_class_add_private(klass, sizeof(C4sDocumentPrivate));

  object_class->dispose = c4s_document_dispose;
  object_class->finalize = c4s_document_finalize;
  object_class->set_property = c4s_document_set_property;
  object_class->get_property = c4s_document_get_property;

  klass->title_changed = c4s_document_title_changed;
  klass->path_changed = c4s_document_path_changed;

  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "User preferences",
      C4S_TYPE_PREFERENCES,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_TITLE,
    g_param_spec_string(
      "title",
      "Title",
      "Title of the document",
      "Untitled",
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_PATH,
    g_param_spec_string(
      "path",
      "Path",
      "Path to file on disk",
      NULL,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_LINE_ENDINGS,
    g_param_spec_enum(
      "line-endings",
      "Line endings",
      "Type of line endings",
      C4S_TYPE_DOCUMENT_LINE_ENDINGS,
      C4S_DOCUMENT_LINE_ENDINGS_DOS,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_ENCODING,
    g_param_spec_string(
      "encoding",
      "Encoding",
      "Character set encoding",
      C4S_DOCUMENT_ENCODING_LATIN1,
      G_PARAM_READWRITE
    )
  );

  document_signals[TITLE_CHANGED] = g_signal_new(
    "title-changed",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sDocumentClass, title_changed),
    NULL, NULL,
    c4s_marshal_VOID__STRING,
    G_TYPE_NONE,
    1,
    G_TYPE_STRING
  );

  document_signals[PATH_CHANGED] = g_signal_new(
    "path-changed",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sDocumentClass, path_changed),
    NULL, NULL,
    c4s_marshal_VOID__STRING,
    G_TYPE_NONE,
    1,
    G_TYPE_STRING
  );
}

GType
c4s_document_line_endings_get_type(void)
{
  static GType etype = 0;
  if(etype == 0)
  {
    static const GEnumValue values[] = {
      {
        C4S_DOCUMENT_LINE_ENDINGS_DOS,
        "C4S_DOCUMENT_LINE_ENDINGS_DOS",
        "dos"
      }, {
        C4S_DOCUMENT_LINE_ENDINGS_UNIX,
        "C4S_DOCUMENT_LINE_ENDINGS_UNIX",
        "unix"
      }, {
        C4S_DOCUMENT_LINE_ENDINGS_MAC,
        "C4S_DOCMUENT_LINE_ENDINGS_MAC",
        "mac"
      }, {
        0,
        NULL,
        NULL
      }
    };

    etype = g_enum_register_static("C4sDocumentLineEndings", values);
  }

  return etype;
}

GType
c4s_document_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sDocumentClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_document_class_init,
      NULL,
      NULL,
      sizeof(C4sDocument),
      0,
      (GInstanceInitFunc)c4s_document_init,
      NULL
    };

    our_type = g_type_register_static(
      GTK_TYPE_SOURCE_BUFFER,
      "C4sDocument",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sDocument*
c4s_document_new(C4sPreferences* preferences)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_DOCUMENT,
    "preferences", preferences,
    NULL
  );

  return C4S_DOCUMENT(object);
}

gboolean
c4s_document_load_file(C4sDocument* document,
                       const gchar* filename,
                       GError** error)
{
  C4sDocumentPrivate* priv;
  GIOChannel* channel;
  GError* local_error;
  const gchar* current_encoding;
  gboolean is_ascii;
  gboolean result;
  gchar* title;

  priv = C4S_DOCUMENT_PRIVATE(document);
  channel = g_io_channel_new_file(filename, "r", error);
  if(channel == NULL) return FALSE;

  gtk_source_buffer_begin_not_undoable_action(GTK_SOURCE_BUFFER(document));
  gtk_text_buffer_begin_user_action(GTK_TEXT_BUFFER(document));

#if 0
  /* Decouple text buffer from components so that these don't get
   * hundreds of updates */
  if(priv->param_info != NULL)
    g_object_set(G_OBJECT(priv->param_info), "buffer", NULL, NULL);

  if(priv->custom_list != NULL)
    g_object_set(G_OBJECT(priv->custom_list), "buffer", NULL, NULL);

  if(priv->highlighter != NULL)
    g_object_set(G_OBJECT(priv->highlighter), "buffer", NULL, NULL);
#endif

  current_encoding = priv->encoding;
  local_error = NULL;

  result = c4s_document_load_stream(
    document,
    channel,
    C4S_DOCUMENT_ENCODING_UTF8,
    &is_ascii,
    &local_error
  );
  
  if(result == FALSE)
  {
    if(local_error->domain == g_convert_error_quark())
    {
      g_error_free(local_error);
      local_error = NULL;

      /* Text is not UTF-8, try LATIN1 */
      result = c4s_document_load_stream(
        document,
        channel,
        C4S_DOCUMENT_ENCODING_LATIN1,
	NULL,
        &local_error
      );

      if(result == FALSE)
      {
        if(local_error->domain == g_convert_error_quark())
        {
          g_error_free(local_error);
          local_error = NULL;

          g_set_error(
            error,
            g_convert_error_quark(),
            G_CONVERT_ERROR_NO_CONVERSION,
            _("The file's encoding is either not known or it is binary data.")
          );
        }
      }
      else
      {
        priv->encoding = C4S_DOCUMENT_ENCODING_LATIN1;
      }
    }
  }
  else
  {
    /* Make LATIN1 if it is also valid */
    if(is_ascii == TRUE)
      priv->encoding = C4S_DOCUMENT_ENCODING_LATIN1;
    else
      priv->encoding = C4S_DOCUMENT_ENCODING_UTF8;
  }

  g_io_channel_unref(channel);

  if(local_error != NULL)
  {
    g_assert(result == FALSE);
    g_propagate_error(error, local_error);
  }

#if 0
  /* Recouple */
  if(priv->highlighter != NULL)
    g_object_set(G_OBJECT(priv->highlighter), "buffer", document, NULL);

  if(priv->custom_list != NULL)
    g_object_set(G_OBJECT(priv->custom_list), "buffer", document, NULL);

  if(priv->param_info != NULL)
    g_object_set(G_OBJECT(priv->param_info), "buffer", document, NULL);
#endif

  gtk_text_buffer_end_user_action(GTK_TEXT_BUFFER(document));
  gtk_source_buffer_end_not_undoable_action(GTK_SOURCE_BUFFER(document));
  gtk_text_buffer_set_modified(GTK_TEXT_BUFFER(document), FALSE);

  if(result == TRUE)
  {
    if(current_encoding != priv->encoding)
      g_object_notify(G_OBJECT(document), "encoding");

    title = c4s_document_get_title_from_path(filename);

    g_signal_emit(
      G_OBJECT(document),
      document_signals[TITLE_CHANGED],
      0,
      title
    );

    g_free(title);

    g_signal_emit(
      G_OBJECT(document),
      document_signals[PATH_CHANGED],
      0,
      filename
    );
  }

  return result;
}

gboolean
c4s_document_save_file(C4sDocument* document,
                       const gchar* filename,
                       GError** error)
{
  C4sDocumentPrivate* priv;
  GIOChannel* channel;
  GError* local_error;
  gboolean result;
  gchar* title;

  priv = C4S_DOCUMENT_PRIVATE(document);
  channel = g_io_channel_new_file(filename, "w", error);
  if(channel == NULL) return FALSE;

  local_error = NULL;

  result = c4s_document_save_stream(
    document,
    channel,
    priv->encoding,
    &local_error
  );
  
  if(result == FALSE && priv->encoding == C4S_DOCUMENT_ENCODING_LATIN1)
  {
    g_error_free(local_error);
    local_error = NULL;

    result = c4s_document_save_stream(
      document,
      channel,
      C4S_DOCUMENT_ENCODING_UTF8,
      &local_error
    );
  }

  g_io_channel_unref(channel);

  if(result == TRUE)
  {
    title = c4s_document_get_title_from_path(filename);

    g_signal_emit(
      G_OBJECT(document),
      document_signals[TITLE_CHANGED],
      0,
      title
    );

    g_free(title);

    if(priv->path == NULL || strcmp(filename, priv->path) != 0)
    {
      g_signal_emit(
        G_OBJECT(document),
        document_signals[PATH_CHANGED],
        0,
        filename
      );
    }

    gtk_text_buffer_set_modified(GTK_TEXT_BUFFER(document), FALSE);
  }
  else
  {
    g_propagate_error(error, local_error);
  }

  return result;
}

const gchar*
c4s_document_get_title(C4sDocument* document)
{
  return C4S_DOCUMENT_PRIVATE(document)->title;
}

const gchar*
c4s_document_get_path(C4sDocument* document)
{
  return C4S_DOCUMENT_PRIVATE(document)->path;
}
