/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-documentwindow.h>

typedef struct _C4sDocumentWindowPrivate C4sDocumentWindowPrivate;
struct _C4sDocumentWindowPrivate {
  C4sKeywordList* keywords;
  C4sPreferences* preferences;
  C4sEngineFunctionList* engine_list;

  C4sDocument* document;
  C4sHighlighter* highlighter;
  C4sCustomFunctionList* custom_list;
  C4sParamInfo* param_info;

  GtkWidget* view;
  GtkWidget* function_bar;
};

#define C4S_DOCUMENT_WINDOW_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_DOCUMENT_WINDOW, \
    C4sDocumentWindowPrivate \
  ))

enum {
  PROP_0,

  PROP_KEYWORDS,
  PROP_PREFERENCES,
  PROP_ENGINE_LIST,

  PROP_DOCUMENT,
  PROP_HIGHLIGHTER,
  PROP_CUSTOM_LIST,
  PROP_PARAM_INFO,

  PROP_VIEW,
  PROP_FUNCTION_BAR
};

static GtkVBoxClass* parent_class = NULL;

static void
c4s_document_window_custom_function_selected_cb(C4sFunctionBar* bar,
                                                C4sCustomFunction* function,
                                                gpointer user_data)
{
  c4s_document_window_scroll_to_function(
    C4S_DOCUMENT_WINDOW(user_data),
    function
  );
}

static void
c4s_document_window_notify_color_text_cb(C4sPreferences* preferences,
                                         GParamSpec* pspec,
                                         gpointer user_data)
{
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;

  window = C4S_DOCUMENT_WINDOW(user_data);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  if(priv->view != NULL)
  {
    gtk_widget_modify_text(
      priv->view,
      GTK_STATE_NORMAL,
      c4s_preferences_get_color_text(preferences)
    );

    gtk_widget_modify_text(
      priv->view,
      GTK_STATE_PRELIGHT,
      c4s_preferences_get_color_text(preferences)
    );

    gtk_widget_modify_text(
      priv->view,
      GTK_STATE_ACTIVE,
      c4s_preferences_get_color_text(preferences)
    );
  }
}

static void
c4s_document_window_notify_tab_width_cb(C4sPreferences* preferences,
                                        GParamSpec* pspec,
                                        gpointer user_data)
{
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;

  window = C4S_DOCUMENT_WINDOW(user_data);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  if(priv->view != NULL)
  {
    gtk_source_view_set_tab_width(
      GTK_SOURCE_VIEW(priv->view),
      c4s_preferences_get_tab_width(preferences)
    );
  }
}

static void
c4s_document_window_notify_auto_indentation_cb(C4sPreferences* preferences,
                                               GParamSpec* pspec,
                                               gpointer user_data)
{
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;

  window = C4S_DOCUMENT_WINDOW(user_data);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  if(priv->view != NULL)
  {
    gtk_source_view_set_auto_indent(
      GTK_SOURCE_VIEW(priv->view),
      c4s_preferences_get_auto_indentation(priv->preferences)
    );
  }
}

static void
c4s_document_window_notify_font_cb(C4sPreferences* preferences,
                                   GParamSpec* pspec,
                                   gpointer user_data)
{
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;
  PangoFontDescription* desc;

  window = C4S_DOCUMENT_WINDOW(user_data);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  if(priv->view != NULL)
  {
    desc = c4s_preferences_get_font(priv->preferences);
    gtk_widget_modify_font(priv->view, desc);
  }
}

static void
c4s_document_window_apply_preferences(C4sDocumentWindow* window)
{
  C4sDocumentWindowPrivate* priv;
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  if(priv->preferences != NULL && priv->view != NULL)
  {
    gtk_widget_modify_text(
      priv->view,
      GTK_STATE_NORMAL,
      c4s_preferences_get_color_text(priv->preferences)
    );

    gtk_widget_modify_text(
      priv->view,
      GTK_STATE_PRELIGHT,
      c4s_preferences_get_color_text(priv->preferences)
    );

    gtk_widget_modify_text(
      priv->view,
      GTK_STATE_ACTIVE,
      c4s_preferences_get_color_text(priv->preferences)
    );

    gtk_source_view_set_tab_width(
      GTK_SOURCE_VIEW(priv->view),
      c4s_preferences_get_tab_width(priv->preferences)
    );

    gtk_source_view_set_auto_indent(
      GTK_SOURCE_VIEW(priv->view),
      c4s_preferences_get_auto_indentation(priv->preferences)
    );

    gtk_widget_modify_font(
      priv->view,
      c4s_preferences_get_font(priv->preferences)
    );
  }
}

static void
c4s_document_window_set_preferences(C4sDocumentWindow* window,
                                    C4sPreferences* preferences)
{
  C4sDocumentWindowPrivate* priv;
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  if(priv->preferences != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_document_window_notify_color_text_cb),
      window
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_document_window_notify_tab_width_cb),
      window
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_document_window_notify_auto_indentation_cb),
      window
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_document_window_notify_font_cb),
      window
    );

    g_object_unref(G_OBJECT(priv->preferences));
  }

  priv->preferences = preferences;

  if(priv->document != NULL)
  {
    g_object_set(
      G_OBJECT(priv->document),
      "preferences", preferences,
      NULL
    );
  }

  if(priv->highlighter != NULL)
  {
    g_object_set(
      G_OBJECT(priv->highlighter),
      "preferences", preferences,
      NULL
    );
  }

  if(priv->view != NULL)
  {
    g_object_set(
      G_OBJECT(priv->view),
      "preferences", preferences,
      NULL
    );
  }

  if(preferences != NULL)
  {
    g_object_ref(G_OBJECT(preferences));

    g_signal_connect(
      G_OBJECT(preferences),
      "notify::color-text",
      G_CALLBACK(c4s_document_window_notify_color_text_cb),
      window
    );

    g_signal_connect(
      G_OBJECT(preferences),
      "notify::tab-width",
      G_CALLBACK(c4s_document_window_notify_tab_width_cb),
      window
    );

    g_signal_connect(
      G_OBJECT(preferences),
      "notify::auto-indentation",
      G_CALLBACK(c4s_document_window_notify_auto_indentation_cb),
      window
    );

    g_signal_connect(
      G_OBJECT(preferences),
      "notify::font",
      G_CALLBACK(c4s_document_window_notify_font_cb),
      window
    );

    c4s_document_window_apply_preferences(window);
  }
}

static void
c4s_document_window_init(C4sDocumentWindow* bar)
{
  C4sDocumentWindowPrivate* priv;
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(bar);

  priv->keywords = NULL;
  priv->preferences = NULL;
  priv->engine_list = NULL;

  priv->document = NULL;
  priv->highlighter = NULL;
  priv->custom_list = NULL;
  priv->param_info = NULL;

  priv->view = NULL;
  priv->function_bar = NULL;
}

static GObject*
c4s_document_window_constructor(GType type,
                                guint n_construct_properties,
                                GObjectConstructParam* construct_param)
{
  GObject* g_object;
  GtkWidget* scroll;
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;

  g_object = G_OBJECT_CLASS(parent_class)->constructor(
    type,
    n_construct_properties,
    construct_param
  );

  window = C4S_DOCUMENT_WINDOW(g_object);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  /* Document might have been set as construct only property */
  if(priv->document == NULL)
    priv->document = c4s_document_new(priv->preferences);

  priv->highlighter = c4s_highlighter_new(
    GTK_TEXT_BUFFER(priv->document),
    priv->keywords,
    priv->engine_list,
    priv->preferences
  );

  priv->custom_list = c4s_custom_function_list_new(
    GTK_TEXT_BUFFER(priv->document),
    priv->highlighter
  );

  priv->param_info = c4s_param_info_new(
    GTK_TEXT_BUFFER(priv->document),
    priv->highlighter,
    priv->engine_list,
    priv->custom_list
  );

  priv->function_bar = c4s_function_bar_new(
    priv->engine_list,
    priv->custom_list,
    priv->param_info
  );

  gtk_widget_show(priv->function_bar);

  g_signal_connect_after(
    G_OBJECT(priv->function_bar),
    "custom-function-selected",
    G_CALLBACK(c4s_document_window_custom_function_selected_cb),
    window
  );

  priv->view = c4s_script_view_new(
    GTK_SOURCE_BUFFER(priv->document),
    priv->param_info,
    priv->preferences
  );

  c4s_document_window_apply_preferences(window);

  gtk_source_view_set_show_line_numbers(GTK_SOURCE_VIEW(priv->view), TRUE);
  gtk_source_view_set_show_right_margin(GTK_SOURCE_VIEW(priv->view), TRUE);
  gtk_widget_show(priv->view);

  scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(
    GTK_SCROLLED_WINDOW(scroll),
    GTK_POLICY_AUTOMATIC,
    GTK_POLICY_AUTOMATIC
  );

  gtk_scrolled_window_set_shadow_type(
    GTK_SCROLLED_WINDOW(scroll),
    GTK_SHADOW_IN
  );

  gtk_container_add(GTK_CONTAINER(scroll), priv->view);
  gtk_widget_show(scroll);

  gtk_box_pack_start(GTK_BOX(window), priv->function_bar, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(window), scroll, TRUE, TRUE, 0);
  gtk_box_set_spacing(GTK_BOX(window), 4);

  return g_object;
}

static void
c4s_document_window_dispose(GObject* object)
{
  C4sDocumentWindow* document_window;
  C4sDocumentWindowPrivate* priv;

  document_window = C4S_DOCUMENT_WINDOW(object);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(document_window);

  if(priv->param_info != NULL)
  {
    g_object_unref(G_OBJECT(priv->param_info));
    priv->param_info = NULL;
  }

  if(priv->custom_list != NULL)
  {
    g_object_unref(G_OBJECT(priv->custom_list));
    priv->custom_list = NULL;
  }

  if(priv->highlighter != NULL)
  {
    g_object_unref(G_OBJECT(priv->highlighter));
    priv->highlighter = NULL;
  }

  if(priv->document != NULL)
  {
    g_object_unref(G_OBJECT(priv->document));
    priv->document = NULL;
  }

  if(priv->engine_list != NULL)
  {
    g_object_unref(G_OBJECT(priv->engine_list));
    priv->engine_list = NULL;
  }

  c4s_document_window_set_preferences(document_window, NULL);

  if(priv->keywords != NULL)
  {
    g_object_unref(G_OBJECT(priv->keywords));
    priv->keywords = NULL;
  }

  if(G_OBJECT_CLASS(parent_class)->dispose)
    G_OBJECT_CLASS(parent_class)->dispose(object);
}

static void
c4s_document_window_set_property(GObject* object,
                                 guint prop_id,
                                 const GValue* value,
                                 GParamSpec* pspec)
{
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;

  g_return_if_fail(C4S_IS_DOCUMENT_WINDOW(object));

  window = C4S_DOCUMENT_WINDOW(object);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  switch(prop_id)
  {
  case PROP_KEYWORDS:
    if(priv->keywords != NULL) g_object_unref(G_OBJECT(priv->keywords));
    priv->keywords = C4S_KEYWORD_LIST(g_value_dup_object(value));

    if(priv->highlighter != NULL)
    {
      g_object_set(
        G_OBJECT(priv->highlighter),
        "keywords", priv->keywords,
        NULL
      );
    }

    break;
  case PROP_PREFERENCES:
    c4s_document_window_set_preferences(
      window,
      C4S_PREFERENCES(g_value_get_object(value))
    );

    break;
  case PROP_ENGINE_LIST:
    if(priv->engine_list != NULL) g_object_unref(G_OBJECT(priv->engine_list));
    priv->engine_list = C4S_ENGINE_FUNCTION_LIST(g_value_dup_object(value));

    if(priv->function_bar != NULL)
    {
      g_object_set(
        G_OBJECT(priv->view),
        "engine-function-list", priv->engine_list,
        NULL
      );
    }

    if(priv->highlighter != NULL)
    {
      g_object_set(
        G_OBJECT(priv->highlighter),
        "engine-function-list", priv->engine_list,
        NULL
      );
    }

    if(priv->param_info != NULL)
    {
      g_object_set(
        G_OBJECT(priv->param_info),
        "param-info", priv->param_info,
        NULL
      );
    }

    break;
  case PROP_DOCUMENT:
    /* This is a construct-only property. We just set the document here,
     * the corresponding highlighter, custom function list and param info
     * are genereted in _constructor */

    /* TODO: Make it a normal read/write/construct property */
    g_assert(priv->document == NULL);
    priv->document = C4S_DOCUMENT(g_value_dup_object(value));
    break;
  case PROP_HIGHLIGHTER:
  case PROP_CUSTOM_LIST:
  case PROP_PARAM_INFO:
  case PROP_VIEW:
  case PROP_FUNCTION_BAR:
    /* Fallthrough, These are non-writable */
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_document_window_get_property(GObject* object,
                                 guint prop_id,
                                 G_GNUC_UNUSED GValue* value,
                                 GParamSpec* pspec)
{
  C4sDocumentWindow* window;
  C4sDocumentWindowPrivate* priv;

  g_return_if_fail(C4S_IS_DOCUMENT_WINDOW(object));

  window = C4S_DOCUMENT_WINDOW(object);
  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  switch(prop_id)
  {
  case PROP_KEYWORDS:
    g_value_set_object(value, G_OBJECT(priv->keywords));
    break;
  case PROP_PREFERENCES:
    g_value_set_object(value, G_OBJECT(priv->preferences));
    break;
  case PROP_ENGINE_LIST:
    g_value_set_object(value, G_OBJECT(priv->engine_list));
    break;
  case PROP_DOCUMENT:
    g_value_set_object(value, G_OBJECT(priv->document));
    break;
  case PROP_HIGHLIGHTER:
    g_value_set_object(value, G_OBJECT(priv->highlighter));
    break;
  case PROP_CUSTOM_LIST:
    g_value_set_object(value, G_OBJECT(priv->custom_list));
    break;
  case PROP_PARAM_INFO:
    g_value_set_object(value, G_OBJECT(priv->param_info));
    break;
  case PROP_VIEW:
    g_value_set_object(value, G_OBJECT(priv->view));
    break;
  case PROP_FUNCTION_BAR:
    g_value_set_object(value, G_OBJECT(priv->function_bar));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_document_window_class_init(C4sDocumentWindowClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = GTK_VBOX_CLASS(g_type_class_peek_parent(klass));

  g_type_class_add_private(klass, sizeof(C4sDocumentWindowPrivate));

  object_class->constructor = c4s_document_window_constructor;
  object_class->dispose = c4s_document_window_dispose;
  object_class->set_property = c4s_document_window_set_property;
  object_class->get_property = c4s_document_window_get_property;

  g_object_class_install_property(
    object_class,
    PROP_KEYWORDS,
    g_param_spec_object(
      "keywords",
      "C4Script keywords",
      "List of keywords to highlight",
      C4S_TYPE_KEYWORD_LIST,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "User preferences",
      C4S_TYPE_PREFERENCES,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_ENGINE_LIST,
    g_param_spec_object(
      "engine-function-list",
      "Engine function list",
      "Engine function list to get engine function definitions from",
      C4S_TYPE_ENGINE_FUNCTION_LIST,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_DOCUMENT,
    g_param_spec_object(
      "document",
      "Document",
      "The shown document",
      C4S_TYPE_DOCUMENT,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_HIGHLIGHTER,
    g_param_spec_object(
      "highlighter",
      "Highlighter",
      "Syntax highlighter",
      C4S_TYPE_HIGHLIGHTER,
      G_PARAM_READABLE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_CUSTOM_LIST,
    g_param_spec_object(
      "custom-function-list",
      "Custom function list",
      "List of defined script functions within the document",
      C4S_TYPE_CUSTOM_FUNCTION_LIST,
      G_PARAM_READABLE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_PARAM_INFO,
    g_param_spec_object(
      "param-info",
      "Param info",
      "Param info object that provides information about the current function",
      C4S_TYPE_PARAM_INFO,
      G_PARAM_READABLE
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_VIEW,
    g_param_spec_object(
      "script-view",
      "Script view",
      "The C4sScriptView widget the document is shown in",
      C4S_TYPE_SCRIPT_VIEW,
      G_PARAM_READABLE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_FUNCTION_BAR,
    g_param_spec_object(
      "function-bar",
      "Function bar",
      "The C4sFunctionBar widget shown on top of the script view",
      C4S_TYPE_FUNCTION_BAR,
      G_PARAM_READABLE
    )
  );
}

GType
c4s_document_window_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sDocumentWindowClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_document_window_class_init,
      NULL,
      NULL,
      sizeof(C4sDocumentWindow),
      0,
      (GInstanceInitFunc)c4s_document_window_init,
      NULL
    };

    our_type = g_type_register_static(
      GTK_TYPE_VBOX,
      "C4sDocumentWindow",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sDocumentWindow*
c4s_document_window_new(C4sKeywordList* keywords,
                        C4sEngineFunctionList* engine_list,
                        C4sPreferences* preferences)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_DOCUMENT_WINDOW,
    "keywords", keywords,
    "engine-function-list", engine_list,
    "preferences", preferences,
    NULL
  );

  return C4S_DOCUMENT_WINDOW(object);
}

C4sDocumentWindow*
c4s_document_window_new_with_document(C4sDocument* document,
                                      C4sKeywordList* keywords,
                                      C4sEngineFunctionList* engine_list,
                                      C4sPreferences* preferences)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_DOCUMENT_WINDOW,
    "document", document,
    "keywords", keywords,
    "engine-function-list", engine_list,
    "preferences", preferences,
    NULL
  );

  return C4S_DOCUMENT_WINDOW(object);
}

C4sDocument*
c4s_document_window_get_document(C4sDocumentWindow* window)
{
  return C4S_DOCUMENT_WINDOW_PRIVATE(window)->document;
}

C4sFunctionBar*
c4s_document_window_get_function_bar(C4sDocumentWindow* window)
{
  return C4S_FUNCTION_BAR(C4S_DOCUMENT_WINDOW_PRIVATE(window)->function_bar);
}

C4sScriptView*
c4s_document_window_get_script_view(C4sDocumentWindow* window)
{
  return C4S_SCRIPT_VIEW(C4S_DOCUMENT_WINDOW_PRIVATE(window)->view);
}

C4sCustomFunctionList*
c4s_document_window_get_custom_function_list(C4sDocumentWindow* window)
{
  return C4S_DOCUMENT_WINDOW_PRIVATE(window)->custom_list;
}

C4sParamInfo*
c4s_document_window_get_param_info(C4sDocumentWindow* window)
{
  return C4S_DOCUMENT_WINDOW_PRIVATE(window)->param_info;
}

void
c4s_document_window_scroll_to_function(C4sDocumentWindow* window,
                                       C4sCustomFunction* function)
{
  C4sDocumentWindowPrivate* priv;
  GtkTextIter begin;
  GtkTextIter end;

  priv = C4S_DOCUMENT_WINDOW_PRIVATE(window);

  c4s_custom_function_get_begin(function, &begin);
  c4s_custom_function_get_end(function, &end);

  gtk_text_buffer_select_range(
    GTK_TEXT_BUFFER(priv->document),
    &begin,
    &end
  );

  gtk_text_view_scroll_to_iter(
    GTK_TEXT_VIEW(priv->view),
    &begin,
    0.1,
    TRUE,
    0.1,
    0.1
  );

  gtk_widget_grab_focus(priv->view);
}
