/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-functionbar.h>
#include <c4scripter/c4s-iconmanager.h>
#include <c4scripter/c4s-marshal.h>

#include <gtk/gtktreemodelsort.h>
#include <gtk/gtkcombobox.h>
#include <gtk/gtkcelllayout.h>
#include <gtk/gtkcellrendererpixbuf.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrenderer.h>

typedef struct _C4sFunctionBarPrivate C4sFunctionBarPrivate;
struct _C4sFunctionBarPrivate {
  C4sEngineFunctionList* engine_list;
  C4sCustomFunctionList* custom_list;
  C4sParamInfo* param_info;

  GtkListStore* custom_store;
  GtkTreeModel* custom_sort;
  GtkWidget* custom_box;
};

#define C4S_FUNCTION_BAR_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_FUNCTION_BAR, \
    C4sFunctionBarPrivate \
  ))

enum {
  PROP_0,

  /* Construct only */
  PROP_ENGINE_LIST,
  PROP_CUSTOM_LIST,
  PROP_PARAM_INFO
};

enum {
  CUSTOM_FUNCTION_SELECTED,
  
  LAST_SIGNAL
};

static GtkHBoxClass* parent_class = NULL;
static guint function_bar_signals[LAST_SIGNAL];

static GdkPixbuf*
c4s_function_bar_get_access_icon(GtkWidget* widget,
                                 C4sCustomFunctionAccess access)
{
  const gchar* stock_id;
  GdkPixbuf* pixbuf;

  switch(access)
  {
  case C4S_CUSTOM_FUNCTION_ACCESS_GLOBAL:
    stock_id = C4S_STOCK_FUNC_GLOBAL;
    break;
  case C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC:
    stock_id = C4S_STOCK_FUNC_PUBLIC;
    break;
  case C4S_CUSTOM_FUNCTION_ACCESS_PROTECTED:
    stock_id = C4S_STOCK_FUNC_PROTECTED;
    break;
  case C4S_CUSTOM_FUNCTION_ACCESS_PRIVATE:
    stock_id = C4S_STOCK_FUNC_PRIVATE;
    break;
  default:
    g_assert_not_reached();
    break;
  }

  pixbuf = gtk_widget_render_icon(widget, stock_id, GTK_ICON_SIZE_MENU, NULL);
  return pixbuf;
}

static gboolean
c4s_function_bar_find_custom_function(C4sFunctionBar* bar,
                                      C4sCustomFunction* function,
				      GtkTreeIter* iter)
{
  C4sFunctionBarPrivate* priv;
  C4sCustomFunction* current_function;
  GtkTreeModel* model;
  gboolean result;

  priv = C4S_FUNCTION_BAR_PRIVATE(bar);
  model = GTK_TREE_MODEL(priv->custom_store);

  for(result = gtk_tree_model_get_iter_first(model, iter);
      result == TRUE;
      result = gtk_tree_model_iter_next(model, iter))
  {
    gtk_tree_model_get(
      model,
      iter,
      C4S_FUNCTION_BAR_COLUMN_FUNCTION,
      &current_function,
      -1
    );

    if(current_function == function)
      return TRUE;
  }

  return FALSE;
}

static void
c4s_function_bar_custom_function_access_changed_cb(C4sCustomFunction* function,
                                                   C4sCustomFunctionAccess acc,
						   gpointer user_data)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;
  GtkTreeIter iter;
  GdkPixbuf* icon;
  gboolean found_function;

  bar = C4S_FUNCTION_BAR(user_data);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  found_function = c4s_function_bar_find_custom_function(bar, function, &iter);
  g_assert(found_function == TRUE);

  icon = c4s_function_bar_get_access_icon(GTK_WIDGET(bar), acc);

  gtk_list_store_set(
    priv->custom_store,
    &iter,
    C4S_FUNCTION_BAR_COLUMN_ACCESS,
    icon,
    -1
  );

  g_object_unref(G_OBJECT(icon));
}

static void
c4s_function_bar_custom_function_name_changed_cb(C4sCustomFunction* function,
                                                 const gchar* name,
						 gpointer user_data)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;
  GtkTreeIter iter;
  gboolean found_function;

  bar = C4S_FUNCTION_BAR(user_data);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  found_function = c4s_function_bar_find_custom_function(bar, function, &iter);
  g_assert(found_function == TRUE);

  gtk_list_store_set(
    priv->custom_store,
    &iter,
    C4S_FUNCTION_BAR_COLUMN_NAME,
    name,
    -1
  );
}

/* This function just unregisters a C4sCustomFunction from the FunctionBar
 * without removing it from the GtkListStore. The function
 * c4s_function_bar_remove_custom_function also removes it from the store. */
static void
c4s_function_bar_deinit_custom_function(C4sFunctionBar* bar,
                                        C4sCustomFunction* function)
{
  g_signal_handlers_disconnect_by_func(
    G_OBJECT(function),
    G_CALLBACK(c4s_function_bar_custom_function_name_changed_cb),
    bar
  );

  g_signal_handlers_disconnect_by_func(
    G_OBJECT(function),
    G_CALLBACK(c4s_function_bar_custom_function_access_changed_cb),
    bar
  );
}

static void
c4s_function_bar_add_custom_function(C4sFunctionBar* bar,
                                     C4sCustomFunction* function)
{
  C4sFunctionBarPrivate* priv;
  GdkPixbuf* icon;
  GtkTreeIter iter;

  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  icon = c4s_function_bar_get_access_icon(
    GTK_WIDGET(bar),
    c4s_custom_function_get_access(function)
  );

  gtk_list_store_append(priv->custom_store, &iter);
  gtk_list_store_set(
    priv->custom_store,
    &iter,
    C4S_FUNCTION_BAR_COLUMN_ACCESS,
    icon,
    C4S_FUNCTION_BAR_COLUMN_NAME,
    c4s_function_get_name(C4S_FUNCTION(function)),
    C4S_FUNCTION_BAR_COLUMN_FUNCTION,
    function,
    -1
  );

  g_object_unref(G_OBJECT(icon));

  g_signal_connect_after(
    G_OBJECT(function),
    "name-changed",
    G_CALLBACK(c4s_function_bar_custom_function_name_changed_cb),
    bar
  );

  g_signal_connect_after(
    G_OBJECT(function),
    "access-changed",
    G_CALLBACK(c4s_function_bar_custom_function_access_changed_cb),
    bar
  );
}

static void
c4s_function_bar_remove_custom_function(C4sFunctionBar* bar,
                                        C4sCustomFunction* function)
{
  C4sFunctionBarPrivate* priv;
  gboolean found_function;
  GtkTreeIter iter;

  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  c4s_function_bar_deinit_custom_function(bar, function);

  found_function = c4s_function_bar_find_custom_function(
    bar,
    function,
    &iter
  );

  g_assert(found_function == TRUE);
  gtk_list_store_remove(priv->custom_store, &iter);
}

static void
c4s_function_bar_custom_box_changed_cb(GtkComboBox* box,
                                       gpointer user_data)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;
  GtkTreeIter iter;
  gboolean has_active_iter;
  C4sCustomFunction* function;

  bar = C4S_FUNCTION_BAR(user_data);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  has_active_iter = gtk_combo_box_get_active_iter(box, &iter);

  /* Normally, this should always be TRUE here, but some recent GTK+
   * versions seem to emit the signal when removing items from the combo box
   * although the actual selected item remains unchanged. */
  /*g_assert(has_active_iter == TRUE);*/
  if(has_active_iter == TRUE)
  {
    gtk_tree_model_get(
      GTK_TREE_MODEL(priv->custom_sort),
      &iter,
      C4S_FUNCTION_BAR_COLUMN_FUNCTION,
      &function,
      -1
    );

    g_signal_emit(
      G_OBJECT(bar),
      function_bar_signals[CUSTOM_FUNCTION_SELECTED],
      0,
      function
    );
  }
}

static void
c4s_function_bar_engine_list_reloaded_cb(G_GNUC_UNUSED
                                         C4sEngineFunctionList* list,
                                         G_GNUC_UNUSED gpointer user_data)
{
}

static void
c4s_function_bar_custom_list_function_added_cb(G_GNUC_UNUSED
                                               C4sCustomFunctionList* list,
                                               C4sCustomFunction* function,
                                               gpointer user_data)
{
  C4sFunctionBar* bar;
  bar = C4S_FUNCTION_BAR(user_data);

  c4s_function_bar_add_custom_function(bar, function);
}

static void
c4s_function_bar_custom_list_function_removed_cb(G_GNUC_UNUSED
                                                 C4sCustomFunctionList* list,
                                                 C4sCustomFunction* function,
						 gpointer user_data)
{
  C4sFunctionBar* bar;
  bar = C4S_FUNCTION_BAR(user_data);

  c4s_function_bar_remove_custom_function(bar, function);
}

static void
c4s_function_bar_param_info_changed_cb(G_GNUC_UNUSED C4sParamInfo* info,
                                       C4sFunction* function,
				       G_GNUC_UNUSED guint parameter,
				       gpointer user_data)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;
  gboolean found_function;
  GtkTreeIter iter;
  GtkTreeIter sort_iter;

  bar = C4S_FUNCTION_BAR(user_data);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  g_signal_handlers_block_by_func(
    G_OBJECT(priv->custom_box),
    G_CALLBACK(c4s_function_bar_custom_box_changed_cb),
    bar
  );

  found_function = FALSE;
  if(function != NULL && C4S_IS_CUSTOM_FUNCTION(function))
  {
    found_function = c4s_function_bar_find_custom_function(
      bar,
      C4S_CUSTOM_FUNCTION(function),
      &iter
    );
  }

  if(found_function == TRUE)
  {
    gtk_tree_model_sort_convert_child_iter_to_iter(
      GTK_TREE_MODEL_SORT(priv->custom_sort),
      &sort_iter,
      &iter
    );

    gtk_combo_box_set_active_iter(
      GTK_COMBO_BOX(priv->custom_box),
      &sort_iter
    );
  }
  else
  {
    gtk_combo_box_set_active(GTK_COMBO_BOX(priv->custom_box), -1);
  }

  g_signal_handlers_unblock_by_func(
    G_OBJECT(priv->custom_box),
    G_CALLBACK(c4s_function_bar_custom_box_changed_cb),
    bar
  );
}

static void
c4s_function_bar_init_custom_functions_cb(C4sCustomFunction* function,
                                          gpointer user_data)
{
  C4sFunctionBar* bar;
  bar = C4S_FUNCTION_BAR(user_data);

  c4s_function_bar_add_custom_function(bar, function);
}

static void
c4s_function_bar_init_custom_functions(C4sFunctionBar* bar)
{
  C4sFunctionBarPrivate* priv;
  C4sCustomFunction* function;
  GtkTreeModel* model;
  GtkTreeIter iter;
  gboolean result;

  priv = C4S_FUNCTION_BAR_PRIVATE(bar);
  model = GTK_TREE_MODEL(priv->custom_store);

  /* Unregister signal handlers of stored custom functions before clearing
   * the list. */
  for(result = gtk_tree_model_get_iter_first(model, &iter);
      result == TRUE;
      result = gtk_tree_model_iter_next(model, &iter))
  {
    gtk_tree_model_get(
      model,
      &iter,
      C4S_FUNCTION_BAR_COLUMN_FUNCTION,
      &function,
      -1
    );

    c4s_function_bar_deinit_custom_function(bar, function);
  }

  gtk_list_store_clear(priv->custom_store);

  /* Reload functions from C4sCustomFunctionList, if available */
  if(priv->custom_list != NULL)
  {
    c4s_custom_function_list_foreach(
      priv->custom_list,
      c4s_function_bar_init_custom_functions_cb,
      bar
    );
  }
}

static void
c4s_function_bar_set_engine_list(C4sFunctionBar* bar,
                                 C4sEngineFunctionList* list)
{
  C4sFunctionBarPrivate* priv;
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  if(priv->engine_list != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->engine_list),
      G_CALLBACK(c4s_function_bar_engine_list_reloaded_cb),
      bar
    );

    g_object_unref(G_OBJECT(priv->engine_list));
  }

  if(list != NULL)
  {
    g_signal_connect_after(
      G_OBJECT(list),
      "reloaded",
      G_CALLBACK(c4s_function_bar_engine_list_reloaded_cb),
      bar
    );

    g_object_ref(G_OBJECT(list));
  }

  priv->engine_list = list;
}

static void
c4s_function_bar_set_custom_list(C4sFunctionBar* bar,
                                 C4sCustomFunctionList* list)
{
  C4sFunctionBarPrivate* priv;
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  if(priv->custom_list != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->custom_list),
      G_CALLBACK(c4s_function_bar_custom_list_function_added_cb),
      bar
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->custom_list),
      G_CALLBACK(c4s_function_bar_custom_list_function_removed_cb),
      bar
    );

    g_object_unref(G_OBJECT(priv->custom_list));
  }

  if(list != NULL)
  {
    g_signal_connect_after(
      G_OBJECT(list),
      "function-added",
      G_CALLBACK(c4s_function_bar_custom_list_function_added_cb),
      bar
    );

    g_signal_connect_after(
      G_OBJECT(list),
      "function-removed",
      G_CALLBACK(c4s_function_bar_custom_list_function_removed_cb),
      bar
    );

    g_object_ref(G_OBJECT(list));
  }

  priv->custom_list = list;

  /* Clear list store, reload functions from new custom list */
  c4s_function_bar_init_custom_functions(bar);
}

static void
c4s_function_bar_set_param_info(C4sFunctionBar* bar,
                                C4sParamInfo* info)
{
  C4sFunctionBarPrivate* priv;
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  if(priv->param_info != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->param_info),
      G_CALLBACK(c4s_function_bar_param_info_changed_cb),
      bar
    );

    g_object_unref(G_OBJECT(priv->param_info));
  }

  if(info != NULL)
  {
    g_signal_connect_after(
      G_OBJECT(info),
      "changed",
      G_CALLBACK(c4s_function_bar_param_info_changed_cb),
      bar
    );

    g_object_ref(G_OBJECT(info));
  }

  priv->param_info = info;
}

static void
c4s_function_bar_init(C4sFunctionBar* bar)
{
  C4sFunctionBarPrivate* priv;
  GtkCellRenderer* custom_renderer_access;
  GtkCellRenderer* custom_renderer_name;

  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  priv->engine_list = NULL;
  priv->custom_list = NULL;
  priv->param_info = NULL;

  priv->custom_store = gtk_list_store_new(
    C4S_FUNCTION_BAR_N_COLUMNS,
    GDK_TYPE_PIXBUF,
    G_TYPE_STRING,
    C4S_TYPE_FUNCTION
  );

  priv->custom_sort = gtk_tree_model_sort_new_with_model(
    GTK_TREE_MODEL(priv->custom_store)
  );

  priv->custom_box = gtk_combo_box_new_with_model(priv->custom_sort);

  gtk_tree_sortable_set_sort_column_id(
    GTK_TREE_SORTABLE(priv->custom_sort),
    C4S_FUNCTION_BAR_COLUMN_NAME,
    GTK_SORT_ASCENDING
  );

  custom_renderer_access = gtk_cell_renderer_pixbuf_new();
  custom_renderer_name = gtk_cell_renderer_text_new();

  gtk_cell_layout_pack_start(
    GTK_CELL_LAYOUT(priv->custom_box),
    custom_renderer_access,
    FALSE
  );

  gtk_cell_layout_pack_start(
    GTK_CELL_LAYOUT(priv->custom_box),
    custom_renderer_name,
    TRUE
  );

  gtk_cell_layout_add_attribute(
    GTK_CELL_LAYOUT(priv->custom_box),
    custom_renderer_access,
    "pixbuf",
    C4S_FUNCTION_BAR_COLUMN_ACCESS
  );

  gtk_cell_layout_add_attribute(
    GTK_CELL_LAYOUT(priv->custom_box),
    custom_renderer_name,
    "text",
    C4S_FUNCTION_BAR_COLUMN_NAME
  );

  g_object_set(
    G_OBJECT(custom_renderer_name),
    "ypad",
    0,
    "ellipsize",
    PANGO_ELLIPSIZE_END,
    "width-chars",
    35,
    NULL
  );

  g_signal_connect_after(
    G_OBJECT(priv->custom_box),
    "changed",
    G_CALLBACK(c4s_function_bar_custom_box_changed_cb),
    bar
  );

  gtk_widget_show(priv->custom_box);
  gtk_box_pack_start(GTK_BOX(bar), priv->custom_box, FALSE, FALSE, 0);
  gtk_box_set_spacing(GTK_BOX(bar), 10);
}

static void 
c4s_function_bar_finalize(GObject* object)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;

  bar = C4S_FUNCTION_BAR(object);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  c4s_function_bar_set_engine_list(bar, NULL);
  c4s_function_bar_set_custom_list(bar, NULL);
  c4s_function_bar_set_param_info(bar, NULL);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_function_bar_set_property(GObject* object,
                              guint prop_id,
                              const GValue* value,
                              GParamSpec* pspec)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;

  g_return_if_fail(C4S_IS_FUNCTION_BAR(object));

  bar = C4S_FUNCTION_BAR(object);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  switch(prop_id)
  {
  case PROP_ENGINE_LIST:
    c4s_function_bar_set_engine_list(
      bar,
      C4S_ENGINE_FUNCTION_LIST(g_value_get_object(value))
    );

    break;
  case PROP_CUSTOM_LIST:
    c4s_function_bar_set_custom_list(
      bar,
      C4S_CUSTOM_FUNCTION_LIST(g_value_get_object(value))
    );

    break;
  case PROP_PARAM_INFO:
    c4s_function_bar_set_param_info(
      bar,
      C4S_PARAM_INFO(g_value_get_object(value))
    );

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_function_bar_get_property(GObject* object,
                              guint prop_id,
                              G_GNUC_UNUSED GValue* value,
                              GParamSpec* pspec)
{
  C4sFunctionBar* bar;
  C4sFunctionBarPrivate* priv;

  g_return_if_fail(C4S_IS_FUNCTION_BAR(object));

  bar = C4S_FUNCTION_BAR(object);
  priv = C4S_FUNCTION_BAR_PRIVATE(bar);

  switch(prop_id)
  {
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_function_bar_class_init(C4sFunctionBarClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = GTK_HBOX_CLASS(g_type_class_peek_parent(klass));

  g_type_class_add_private(klass, sizeof(C4sFunctionBarPrivate));

  object_class->finalize = c4s_function_bar_finalize;
  object_class->set_property = c4s_function_bar_set_property;
  object_class->get_property = c4s_function_bar_get_property;

  g_object_class_install_property(
    object_class,
    PROP_ENGINE_LIST,
    g_param_spec_object(
      "engine-function-list",
      "Engine function list",
      "Engine function list to get engine function definitions from",
      C4S_TYPE_ENGINE_FUNCTION_LIST,
      G_PARAM_READWRITE
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_CUSTOM_LIST,
    g_param_spec_object(
      "custom-function-list",
      "Custom function list",
      "Custom function list to get custom function definitions from",
      C4S_TYPE_CUSTOM_FUNCTION_LIST,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_PARAM_INFO,
    g_param_spec_object(
      "param-info",
      "Param info",
      "Param info object to get current function from",
      C4S_TYPE_PARAM_INFO,
      G_PARAM_READWRITE
    )
  );
  
  function_bar_signals[CUSTOM_FUNCTION_SELECTED] = g_signal_new(
    "custom-function-selected",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    0,
    NULL, NULL,
    c4s_marshal_VOID__OBJECT,
    G_TYPE_NONE,
    1,
    C4S_TYPE_CUSTOM_FUNCTION
  );
}

GType
c4s_function_bar_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sFunctionBarClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_function_bar_class_init,
      NULL,
      NULL,
      sizeof(C4sFunctionBar),
      0,
      (GInstanceInitFunc)c4s_function_bar_init,
      NULL
    };

    our_type = g_type_register_static(
      GTK_TYPE_HBOX,
      "C4sFunctionBar",
      &our_info,
      0
    );
  }

  return our_type;
}

GtkWidget*
c4s_function_bar_new(C4sEngineFunctionList* engine_list,
                     C4sCustomFunctionList* custom_list,
		     C4sParamInfo* param_info)
{
  GObject* object;

  /* TODO: parent as construct only property, set in _constructor */
  object = g_object_new(
    C4S_TYPE_FUNCTION_BAR,
    "engine-function-list", engine_list,
    "custom-function-list", custom_list,
    "param-info", param_info,
    NULL
  );

  return GTK_WIDGET(object);
}

GtkListStore*
c4s_function_bar_get_list_store(C4sFunctionBar* bar)
{
  return C4S_FUNCTION_BAR_PRIVATE(bar)->custom_store;
}
