/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-paraminfo.h>
#include <c4scripter/c4s-textutil.h>
#include <c4scripter/c4s-marshal.h>

enum {
  PROP_0,

  PROP_BUFFER,
  PROP_HIGHLIGHTER,
  PROP_ENGINE_LIST,
  PROP_CUSTOM_LIST
};

enum {
  CHANGED,

  LAST_SIGNAL
};

typedef struct _C4sParamInfoPrivate C4sParamInfoPrivate;
struct _C4sParamInfoPrivate {
  const C4sFunction* function;
  guint parameter;

  /* Properties */
  GtkTextBuffer* buffer;
  C4sHighlighter* highlighter;
  C4sEngineFunctionList* engine_list;
  C4sCustomFunctionList* custom_list;
};

#define C4S_PARAM_INFO_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_PARAM_INFO, \
    C4sParamInfoPrivate \
  ))

static GObjectClass* parent_class = NULL;
static guint paraminfo_signals[LAST_SIGNAL];

static void
c4s_param_info_changed(C4sParamInfo* info,
                       const C4sFunction* function,
                       guint parameter)
{
  C4sParamInfoPrivate* priv;
  priv = C4S_PARAM_INFO_PRIVATE(info);

  priv->function = function;
  priv->parameter = parameter;
}

static void
c4s_param_info_update(C4sParamInfo* info)
{
  C4sParamInfoPrivate* priv;
  GtkTextIter pos;
  GtkTextIter end;
  gint level;
  const C4sFunction* function;
  guint param;
  gunichar c;
  gchar* text;

  priv = C4S_PARAM_INFO_PRIVATE(info);
  gtk_text_buffer_get_iter_at_mark(
    priv->buffer,
    &pos,
    gtk_text_buffer_get_insert(priv->buffer)
  );

  if(c4s_highlighter_is_comment(priv->highlighter, &pos))
    c4s_highlighter_backward_out_of_comment(priv->highlighter, &pos);
  else if(c4s_highlighter_is_string(priv->highlighter, &pos))
    c4s_highlighter_backward_out_of_string(priv->highlighter, &pos);

  level = 0;
  param = 0;

  /* TODO: I would prefer a loop without all those returns inbetween. */
  while(gtk_text_iter_is_start(&pos) == FALSE)
  {
    if(c4s_highlighter_ends_comment(priv->highlighter, &pos))
    {
      c4s_highlighter_backward_out_of_comment(priv->highlighter, &pos);
    }
    else if(c4s_highlighter_ends_string(priv->highlighter, &pos))
    {
      if(level < 0 && priv->function != NULL)
      {
        g_signal_emit(G_OBJECT(info), paraminfo_signals[CHANGED], 0, NULL, 0);
        return;
      }
      else 
      {
        c4s_highlighter_backward_out_of_string(priv->highlighter, &pos);
      }
    }
    else
    {
      gtk_text_iter_backward_char(&pos);
      c = gtk_text_iter_get_char(&pos);

      if(g_unichar_isspace(c) == FALSE)
      {
        if(level < 0)
        {
          if(g_unichar_isalnum(c) || c == '_')
          {
            end = pos;
            gtk_text_iter_forward_char(&end);
            c4s_backward_identifier(&pos);
            text = gtk_text_buffer_get_slice(priv->buffer, &pos, &end, FALSE);

            function = C4S_FUNCTION(
              c4s_custom_function_list_find_function(
                priv->custom_list,
                text
              )
            );

            if(function == NULL)
            {
              function = C4S_FUNCTION(
                c4s_engine_function_list_find_function(
                  priv->engine_list,
                  text
                )
              );
            }

            /* Guarantee that param is 0 when no function is set. */
            if(function == NULL) param = 0;

            g_free(text);

            g_signal_emit(
              G_OBJECT(info),
              paraminfo_signals[CHANGED],
              0,
              function,
              param
            );

            return;
          }

          /* Was not a function paranthesis. */
          level = 0;
          param = 0;
        }

        if(c == ')')
        {
          ++ level;
        }
        else if(c == '(')
        {
          -- level;
        }
        else if(c == ',' && level == 0)
        {
          ++ param;
        }
        else if(c == ';' || c == '{' || c == '}' || c == ':')
        {
          g_signal_emit(
            G_OBJECT(info),
            paraminfo_signals[CHANGED],
            0,
            NULL,
            0
          );

          return;
        }
      }
    }
  }

  /* Might reach here if a function call is not complete and the
   * beginning of the document has been reached. */
  g_signal_emit(G_OBJECT(info), paraminfo_signals[CHANGED], 0, NULL, 0);
}

static void
c4s_param_info_mark_set_cb(GtkTextBuffer* buffer,
                           G_GNUC_UNUSED GtkTextIter* iter,
                           GtkTextMark* mark,
                           gpointer user_data)
{
  C4sParamInfo* info;
  info = C4S_PARAM_INFO(user_data);

  if(gtk_text_buffer_get_insert(buffer) == mark)
  {
    c4s_param_info_update(info);
  }
}

static void
c4s_param_info_reloaded_cb(G_GNUC_UNUSED C4sEngineFunctionList* list,
                           gpointer user_data)
{
  C4sParamInfo* info;
  info = C4S_PARAM_INFO(user_data);

  c4s_param_info_update(info);
}

static void
c4s_param_info_highlighted_cb(G_GNUC_UNUSED C4sHighlighter* highlighter,
                              G_GNUC_UNUSED GtkTextIter* from,
                              G_GNUC_UNUSED GtkTextIter* to,
                              gpointer user_data)
{
  C4sParamInfo* info;
  info = C4S_PARAM_INFO(user_data);

  c4s_param_info_update(info);
}

static void
c4s_param_info_set_text_buffer(C4sParamInfo* info,
                               GtkTextBuffer* buffer)
{
  C4sParamInfoPrivate* priv;
  priv = C4S_PARAM_INFO_PRIVATE(info);

  if(priv->buffer != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->buffer),
      G_CALLBACK(c4s_param_info_mark_set_cb),
      info
    );

    g_object_unref(G_OBJECT(priv->buffer));
  }

  if(buffer != NULL)
  {
    g_object_ref(G_OBJECT(buffer));

    g_signal_connect_after(
      G_OBJECT(buffer),
      "mark-set",
      G_CALLBACK(c4s_param_info_mark_set_cb),
      info
    );
  }

  priv->buffer = buffer;
}

static void
c4s_param_info_set_highlighter(C4sParamInfo* info,
                               C4sHighlighter* highlighter)
{
  C4sParamInfoPrivate* priv;
  priv = C4S_PARAM_INFO_PRIVATE(info);

  if(priv->highlighter != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->highlighter),
      G_CALLBACK(c4s_param_info_highlighted_cb),
      info
    );

    g_object_unref(G_OBJECT(priv->highlighter));
  }

  if(highlighter != NULL)
  {
    g_object_ref(G_OBJECT(highlighter));

    g_signal_connect_after(
      G_OBJECT(highlighter),
      "highlighted",
      G_CALLBACK(c4s_param_info_highlighted_cb),
      info
    );
  }

  priv->highlighter = highlighter;
}

static void
c4s_param_info_set_engine_function_list(C4sParamInfo* info,
                                        C4sEngineFunctionList* engine_list)
{
  C4sParamInfoPrivate* priv;
  priv = C4S_PARAM_INFO_PRIVATE(info);

  if(priv->engine_list != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->engine_list),
      G_CALLBACK(c4s_param_info_reloaded_cb),
      info
    );

    g_object_unref(G_OBJECT(priv->engine_list));
  }

  if(engine_list != NULL)
  {
    g_object_ref(G_OBJECT(engine_list));

    g_signal_connect_after(
      G_OBJECT(engine_list),
      "reloaded",
      G_CALLBACK(c4s_param_info_reloaded_cb),
      info
    );
  }

  priv->engine_list = engine_list;
}

static void
c4s_param_info_set_custom_function_list(C4sParamInfo* info,
                                        C4sCustomFunctionList* custom_list)
{
  C4sParamInfoPrivate* priv;
  priv = C4S_PARAM_INFO_PRIVATE(info);

  if(priv->custom_list != NULL)
  {
    g_object_unref(G_OBJECT(priv->custom_list));
  }

  if(custom_list != NULL)
  {
    g_object_ref(G_OBJECT(custom_list));
  }

  priv->custom_list = custom_list;
}

static void
c4s_param_info_init(C4sParamInfo* param_info)
{
  C4sParamInfoPrivate* priv;
  priv = C4S_PARAM_INFO_PRIVATE(param_info);

  priv->function = NULL;
  priv->parameter = 0;

  priv->buffer = NULL;
  priv->highlighter = NULL;
  priv->engine_list = NULL;
  priv->custom_list = NULL;
}

static void 
c4s_param_info_finalize(GObject* object)
{
  C4sParamInfo* param_info;
  C4sParamInfoPrivate* priv;

  param_info = C4S_PARAM_INFO(object);
  priv = C4S_PARAM_INFO_PRIVATE(param_info);

  c4s_param_info_set_text_buffer(param_info, NULL);
  c4s_param_info_set_highlighter(param_info, NULL);
  c4s_param_info_set_engine_function_list(param_info, NULL);
  c4s_param_info_set_custom_function_list(param_info, NULL);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_param_info_set_property(GObject* object,
                            guint prop_id,
                            const GValue* value,
                            GParamSpec* pspec)
{
  C4sParamInfo* param_info;
  C4sParamInfoPrivate* priv;

  g_return_if_fail(C4S_IS_PARAM_INFO(object));

  param_info = C4S_PARAM_INFO(object);
  priv = C4S_PARAM_INFO_PRIVATE(param_info);

  switch(prop_id)
  {
  case PROP_BUFFER:
    c4s_param_info_set_text_buffer(
      param_info,
      GTK_TEXT_BUFFER(g_value_get_object(value))
    );

    break;
  case PROP_HIGHLIGHTER:
    c4s_param_info_set_highlighter(
      param_info,
      C4S_HIGHLIGHTER(g_value_get_object(value))
    );

    break;
  case PROP_ENGINE_LIST:
    c4s_param_info_set_engine_function_list(
      param_info,
      C4S_ENGINE_FUNCTION_LIST(g_value_get_object(value))
    );

    break;
  case PROP_CUSTOM_LIST:
    c4s_param_info_set_custom_function_list(
      param_info,
      C4S_CUSTOM_FUNCTION_LIST(g_value_get_object(value))
    );

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_param_info_get_property(GObject* object,
                            guint prop_id,
                            G_GNUC_UNUSED GValue* value,
                            GParamSpec* pspec)
{
  C4sParamInfo* param_info;
  C4sParamInfoPrivate* priv;

  g_return_if_fail(C4S_IS_PARAM_INFO(object));

  param_info = C4S_PARAM_INFO(object);
  priv = C4S_PARAM_INFO_PRIVATE(param_info);

  switch(prop_id)
  {
    /* Those are all construct only */
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_param_info_class_init(C4sParamInfoClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sParamInfoPrivate));

  object_class->finalize = c4s_param_info_finalize;
  object_class->set_property = c4s_param_info_set_property;
  object_class->get_property = c4s_param_info_get_property;
  klass->changed = c4s_param_info_changed;

  g_object_class_install_property(
    object_class,
    PROP_BUFFER,
    g_param_spec_object(
      "buffer",
      "Text buffer",
      "Text buffer that is scanned for current function",
      GTK_TYPE_TEXT_BUFFER,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_HIGHLIGHTER,
    g_param_spec_object(
      "highlighter",
      "Highlighter",
      "Highlighter used to get notification about document changes",
      C4S_TYPE_HIGHLIGHTER,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_ENGINE_LIST,
    g_param_spec_object(
      "engine-function-list",
      "Engine function list",
      "List of engine functions to recognize",
      C4S_TYPE_ENGINE_FUNCTION_LIST,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_CUSTOM_LIST,
    g_param_spec_object(
      "custom-function-list",
      "Custom function list",
      "List of custom functions to recognize",
      C4S_TYPE_CUSTOM_FUNCTION_LIST,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  paraminfo_signals[CHANGED] = g_signal_new(
    "changed",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sParamInfoClass, changed),
    NULL, NULL,
    c4s_marshal_VOID__OBJECT_UINT,
    G_TYPE_NONE,
    2,
    C4S_TYPE_FUNCTION,
    G_TYPE_UINT
  );
}

GType
c4s_param_info_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sParamInfoClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_param_info_class_init,
      NULL,
      NULL,
      sizeof(C4sParamInfo),
      0,
      (GInstanceInitFunc)c4s_param_info_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sParamInfo",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sParamInfo*
c4s_param_info_new(GtkTextBuffer* buffer,
                   C4sHighlighter* highlighter,
                   C4sEngineFunctionList* engine_list,
                   C4sCustomFunctionList* custom_list)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_PARAM_INFO,
    "buffer", buffer,
    "highlighter", highlighter,
    "engine-function-list", engine_list,
    "custom-function-list", custom_list,
    NULL
  );

  return C4S_PARAM_INFO(object);
}

const C4sFunction*
c4s_param_info_get_function(const C4sParamInfo* param_info)
{
  return C4S_PARAM_INFO_PRIVATE(param_info)->function;
}

guint
c4s_param_info_get_parameter(const C4sParamInfo* param_info)
{
  return C4S_PARAM_INFO_PRIVATE(param_info)->parameter;
}
