/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-scriptview.h>

#include <string.h>

/* TODO: Handle color-text, auto-indentation, tab-width, font preferences */

typedef struct _C4sScriptViewPrivate C4sScriptViewPrivate;
struct _C4sScriptViewPrivate {
  C4sParamInfo* param_info;
  C4sPreferences* preferences;

  GdkRectangle tip_rect;
  PangoLayout* layout;
};

#define C4S_SCRIPT_VIEW_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_SCRIPT_VIEW, \
    C4sScriptViewPrivate \
  ))

enum {
  PROP_0,

  PROP_PARAM_INFO,
  PROP_PREFERENCES
};

static GtkSourceViewClass* parent_class = NULL;

static gchar*
c4s_script_view_create_tip_text(const C4sFunction* function,
                               guint parameter,
                               gsize* param_len,
                               gsize* full_len)
{
  GString* str;
  guint i;
  C4sValueType type;
  const C4sFunctionParameter* parameter_obj;
  gsize n_parameters;

  str = g_string_sized_new(128);
  type = c4s_function_get_return_type(function);

  g_string_append(str, c4s_value_type_to_string(type & ~C4S_VALUE_TYPE_REF));
  if(type & C4S_VALUE_TYPE_REF) g_string_append(str, "&amp;");

  g_string_append_c(str, ' ');
  g_string_append(str, c4s_function_get_name(function));
  g_string_append_c(str, '(');

  parameter_obj = c4s_function_get_parameters(function, &n_parameters);
  for(i = 0; i < n_parameters; ++ i)
  {
    if(i > 0) g_string_append(str, ", ");

    if(i == parameter)
    {
      *param_len = str->len;
      g_string_append(str, "<b>");
    }

    type = parameter_obj[i].type;
    g_string_append(str, c4s_value_type_to_string(type & ~C4S_VALUE_TYPE_REF));
    if(type & C4S_VALUE_TYPE_REF) g_string_append(str, "&amp;");

    g_string_append_c(str, ' ');
    g_string_append(str, parameter_obj[i].name);

    if(i == parameter) g_string_append(str, "</b>");
  }

  if(parameter >= n_parameters)
    *param_len = str->len;

  g_string_append_c(str, ')');
  *full_len = str->len;
  return g_string_free(str, FALSE);
}

static void
c4s_script_view_redraw_tip_rect(C4sScriptView* view)
{
  C4sScriptViewPrivate* priv;
  GdkRectangle redraw_rect;
  GdkWindow* window;

  priv = C4S_SCRIPT_VIEW_PRIVATE(view);
  if(priv->tip_rect.width > 0)
  {
    window = gtk_text_view_get_window(
      GTK_TEXT_VIEW(view),
      GTK_TEXT_WINDOW_TEXT
    );

    gtk_text_view_buffer_to_window_coords(
      GTK_TEXT_VIEW(view),
      GTK_TEXT_WINDOW_TEXT,
      priv->tip_rect.x,
      priv->tip_rect.y,
      &redraw_rect.x,
      &redraw_rect.y
    );

    redraw_rect.width = priv->tip_rect.width;
    redraw_rect.height = priv->tip_rect.height;

    gdk_window_invalidate_rect(window, &redraw_rect, FALSE);
  }
}

static void
c4s_script_view_recalc_tip_rect(C4sScriptView* view)
{
  C4sScriptViewPrivate* priv;
  const C4sFunction* function;
  guint parameter;
  gchar* text;
  gsize param_len;
  gsize full_len;
  gint offset_width;
  gint text_width;
  gint text_height;
  GtkTextIter insert_iter;
  GtkTextBuffer* buffer;
  GdkRectangle iter_location;

  priv = C4S_SCRIPT_VIEW_PRIVATE(view);
  function = c4s_param_info_get_function(priv->param_info);

  if(function == NULL)
  {
    priv->tip_rect.x = 0;
    priv->tip_rect.y = 0;
    priv->tip_rect.width = 0;
    priv->tip_rect.height = 0;
  }
  else
  {
    parameter = c4s_param_info_get_parameter(priv->param_info);

    text = c4s_script_view_create_tip_text(
      function,
      parameter,
      &param_len,
      &full_len
    );

    if(priv->layout == NULL)
      priv->layout = gtk_widget_create_pango_layout(GTK_WIDGET(view), NULL);

    pango_layout_set_markup(priv->layout, text, param_len);
    pango_layout_get_pixel_size(priv->layout, &offset_width, &text_height);

    /* TODO: Would be enough to query remaining extent? */
    pango_layout_set_markup(priv->layout, text, full_len);
    pango_layout_get_pixel_size(priv->layout, &text_width, NULL);

    g_free(text);

    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

    gtk_text_buffer_get_iter_at_mark(
      buffer,
      &insert_iter,
      gtk_text_buffer_get_insert(buffer)
    );

    gtk_text_view_get_iter_location(
      GTK_TEXT_VIEW(view),
      &insert_iter,
      &iter_location
    );

    priv->tip_rect.x = iter_location.x - (offset_width + 2);
    priv->tip_rect.y = iter_location.y + iter_location.height;
    priv->tip_rect.width = text_width + 4;
    priv->tip_rect.height = text_height + 4;
  }
}

static void
c4s_script_view_param_info_changed_cb(G_GNUC_UNUSED C4sParamInfo* param_info,
                                      G_GNUC_UNUSED C4sFunction* function,
                                      G_GNUC_UNUSED guint parameter,
                                      gpointer user_data)
{
  C4sScriptView* view;
  C4sScriptViewPrivate* priv;

  view = C4S_SCRIPT_VIEW(user_data);
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  if(c4s_preferences_get_view_tooltips(priv->preferences))
    c4s_script_view_redraw_tip_rect(view);

  c4s_script_view_recalc_tip_rect(view);

  if(c4s_preferences_get_view_tooltips(priv->preferences))
    c4s_script_view_redraw_tip_rect(view);
}

static void
c4s_script_view_preferences_notify_cb(G_GNUC_UNUSED GObject* object,
                                      GParamSpec* spec,
                                      gpointer user_data)
{
  if(strcmp(spec->name, "view-tooltips") == 0)
    c4s_script_view_redraw_tip_rect(C4S_SCRIPT_VIEW(user_data));
}

static void
c4s_script_view_set_param_info(C4sScriptView* view,
                               C4sParamInfo* info)
{
  C4sScriptViewPrivate* priv;
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  if(priv->param_info != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->param_info),
      G_CALLBACK(c4s_script_view_param_info_changed_cb),
      view
    );

    g_object_unref(G_OBJECT(priv->param_info));
  }

  if(info != NULL)
  {
    g_signal_connect_after(
      G_OBJECT(info),
      "changed",
      G_CALLBACK(c4s_script_view_param_info_changed_cb),
      view
    );

    g_object_ref(G_OBJECT(info));
  }

  priv->param_info = info;

  if(GTK_WIDGET_REALIZED(GTK_WIDGET(view)))
    c4s_script_view_redraw_tip_rect(view);
}

static void
c4s_script_view_set_preferences(C4sScriptView* view,
                                C4sPreferences* preferences)
{
  C4sScriptViewPrivate* priv;
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  if(priv->preferences != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_script_view_preferences_notify_cb),
      view
    );

    g_object_unref(G_OBJECT(priv->preferences));
  }

  if(preferences != NULL)
  {
    g_signal_connect_after(
      G_OBJECT(preferences),
      "notify",
      G_CALLBACK(c4s_script_view_preferences_notify_cb),
      view
    );

    g_object_ref(G_OBJECT(preferences));
  }

  priv->preferences = preferences;

  if(GTK_WIDGET_REALIZED(GTK_WIDGET(view)))
    c4s_script_view_redraw_tip_rect(view);
}

static void
c4s_script_view_init(C4sScriptView* view)
{
  C4sScriptViewPrivate* priv;
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  priv->param_info = NULL;
  priv->preferences = NULL;
  priv->layout = NULL;

  priv->tip_rect.x = 0;
  priv->tip_rect.y = 0;
  priv->tip_rect.width = 0;
  priv->tip_rect.height = 0;

  gtk_source_view_set_indent_on_tab(GTK_SOURCE_VIEW(view), TRUE);
}

static void 
c4s_script_view_finalize(GObject* object)
{
  C4sScriptView* view;
  C4sScriptViewPrivate* priv;

  view = C4S_SCRIPT_VIEW(object);
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  if(priv->layout != NULL)
    g_object_unref(G_OBJECT(priv->layout));

  c4s_script_view_set_param_info(view, NULL);
  c4s_script_view_set_preferences(view, NULL);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_script_view_set_property(GObject* object,
                             guint prop_id,
                             const GValue* value,
                             GParamSpec* pspec)
{
  C4sScriptView* view;
  C4sScriptViewPrivate* priv;

  g_return_if_fail(C4S_IS_SCRIPT_VIEW(object));

  view = C4S_SCRIPT_VIEW(object);
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  switch(prop_id)
  {
  case PROP_PARAM_INFO:
    c4s_script_view_set_param_info(
      view,
      C4S_PARAM_INFO(g_value_get_object(value))
    );

    break;
  case PROP_PREFERENCES:
    c4s_script_view_set_preferences(
      view,
      C4S_PREFERENCES(g_value_get_object(value))
    );

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_script_view_get_property(GObject* object,
                             guint prop_id,
                             G_GNUC_UNUSED GValue* value,
                             GParamSpec* pspec)
{
  C4sScriptView* view;
  C4sScriptViewPrivate* priv;

  g_return_if_fail(C4S_IS_SCRIPT_VIEW(object));

  view = C4S_SCRIPT_VIEW(object);
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  switch(prop_id)
  {
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_script_view_style_set(GtkWidget* widget,
                          GtkStyle* previous_style)
{
  C4sScriptView* view;
  C4sScriptViewPrivate* priv;

  view = C4S_SCRIPT_VIEW(widget);
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  if(GTK_WIDGET_CLASS(parent_class)->style_set != NULL)
    GTK_WIDGET_CLASS(parent_class)->style_set(widget, previous_style);

  if(c4s_preferences_get_view_tooltips(priv->preferences))
    c4s_script_view_redraw_tip_rect(view);

  c4s_script_view_recalc_tip_rect(view);

  if(c4s_preferences_get_view_tooltips(priv->preferences))
    c4s_script_view_redraw_tip_rect(view);
}

static gboolean
c4s_script_view_expose_event(GtkWidget* widget,
                             GdkEventExpose* event)
{
  C4sScriptView* view;
  C4sScriptViewPrivate* priv;
  GdkRectangle redraw_rect;
  GdkRectangle intersection;
  gboolean result;
  gboolean event_handled;
  GdkWindow* window;
  cairo_t* cr;

  view = C4S_SCRIPT_VIEW(widget);
  priv = C4S_SCRIPT_VIEW_PRIVATE(view);

  if(GTK_WIDGET_CLASS(parent_class)->expose_event)
  {
    event_handled = GTK_WIDGET_CLASS(parent_class)->expose_event(
      widget,
      event
    );
  }

  window = gtk_text_view_get_window(GTK_TEXT_VIEW(view), GTK_TEXT_WINDOW_TEXT);

  if(event->window != window)
    return event_handled;
  if(c4s_preferences_get_view_tooltips(priv->preferences) == FALSE)
    return event_handled;

  gtk_text_view_buffer_to_window_coords(
    GTK_TEXT_VIEW(view),
    GTK_TEXT_WINDOW_TEXT,
    priv->tip_rect.x,
    priv->tip_rect.y,
    &redraw_rect.x,
    &redraw_rect.y
  );

  redraw_rect.width = priv->tip_rect.width;
  redraw_rect.height = priv->tip_rect.height;

  result = gdk_rectangle_intersect(
    &event->area,
    &redraw_rect,
    &intersection
  );

  if(result == FALSE) return event_handled;

  cr = gdk_cairo_create(window);

  cairo_rectangle(
    cr,
    intersection.x,
    intersection.y,
    intersection.width,
    intersection.height
  );

  cairo_clip(cr);
  cairo_set_line_width(cr, 1.0);

  cairo_rectangle(
    cr,
    redraw_rect.x,
    redraw_rect.y,
    redraw_rect.width,
    redraw_rect.height
  );

  cairo_set_source_rgb(cr, 1.0, 1.0, 192.0/255.0);
  cairo_fill_preserve(cr);
  cairo_set_source_rgb(cr, 0, 0, 0);
  cairo_stroke_preserve(cr);
  cairo_destroy(cr);

  gtk_paint_layout(
    widget->style,
    window,
    GTK_WIDGET_STATE(widget),
    FALSE,
    &intersection,
    widget,
    NULL,
    redraw_rect.x + 2,
    redraw_rect.y + 2,
    priv->layout
  );

  return event_handled;
}

static void
c4s_script_view_class_init(C4sScriptViewClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  GtkWidgetClass* widget_class = GTK_WIDGET_CLASS(klass);
  parent_class = GTK_SOURCE_VIEW_CLASS(g_type_class_peek_parent(klass));

  g_type_class_add_private(klass, sizeof(C4sScriptViewPrivate));

  object_class->finalize = c4s_script_view_finalize;
  object_class->set_property = c4s_script_view_set_property;
  object_class->get_property = c4s_script_view_get_property;

  widget_class->style_set = c4s_script_view_style_set;
  widget_class->expose_event = c4s_script_view_expose_event;

  g_object_class_install_property(
    object_class,
    PROP_PARAM_INFO,
    g_param_spec_object(
      "param-info",
      "Param Info",
      "The C4sParamInfo object to get the current function from",
      C4S_TYPE_PARAM_INFO,
      G_PARAM_READWRITE
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "Preferences to know whether to draw parameter tooltips or not",
      C4S_TYPE_PREFERENCES,
      G_PARAM_READWRITE
    )
  );
}

GType
c4s_script_view_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sScriptViewClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_script_view_class_init,
      NULL,
      NULL,
      sizeof(C4sScriptView),
      0,
      (GInstanceInitFunc)c4s_script_view_init,
      NULL
    };

    our_type = g_type_register_static(
      GTK_TYPE_SOURCE_VIEW,
      "C4sScriptView",
      &our_info,
      0
    );
  }

  return our_type;
}

GtkWidget*
c4s_script_view_new(GtkSourceBuffer* buffer,
                    C4sParamInfo* param_info,
                    C4sPreferences* preferences)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_SCRIPT_VIEW,
    "buffer", buffer,
    "param-info", param_info,
    "preferences", preferences,
    NULL
  );

  return GTK_WIDGET(object);
}
