/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <c4scripter/comm/c4s-comm.h>

#include <glib/giochannel.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>

#include <string.h>

#ifndef UNIX_PATH_MAX
# define UNIX_PATH_MAX 108
#endif

static const gsize RECV_BUFFER_SIZE = 256;

typedef struct _C4sCommConn C4sCommConn;
struct _C4sCommConn
{
  C4sCommBackend* backend;
  GIOChannel* channel;
  gchar buffer[256];
  size_t size;
};

struct _C4sCommBackend
{
  C4sCommFileFunc file_func;
  gpointer user_data;

  GIOChannel* channel;
  struct sockaddr_un addr;
  gboolean host;

  GSList* conns;
};

static GIOChannel*
c4s_comm_unix_make_io_channel(int sock)
{
  GIOChannel* channel;
  channel = g_io_channel_unix_new(sock);
  g_io_channel_set_encoding(channel, NULL, NULL);
  g_io_channel_set_close_on_unref(channel, TRUE);
  return channel;
}

static gboolean
c4s_comm_unix_recv_func(G_GNUC_UNUSED GIOChannel* source,
                        GIOCondition condition,
                        gpointer user_data)
{
  C4sCommConn* conn;
  C4sCommBackend* backend;
  GIOStatus status;
  gsize res;
  char* occ;
  gchar* file;
  
  conn = user_data;
  backend = conn->backend;
  
  if(condition != G_IO_IN)
    goto error;

  status = g_io_channel_read_chars(
    conn->channel,
    conn->buffer + conn->size,
    RECV_BUFFER_SIZE - conn->size,
    &res,
    NULL
  );
  
  if(status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN)
    goto error;

  conn->size += res;

  do
  {
    occ = memchr(conn->buffer, '\n', conn->size);
    
    /* Name too long */
    if(occ == NULL && conn->size == RECV_BUFFER_SIZE)
      goto error;

    if(occ != NULL)
    {
      file = g_strndup(conn->buffer, occ - conn->buffer);
      memmove(conn->buffer, occ + 1, conn->size - (occ - conn->buffer) + 1);
      conn->size -= (occ - conn->buffer + 1);
      backend->file_func(file, backend->user_data);
      g_free(file);
    }
  } while(occ != NULL);

  return TRUE;
error:
  g_io_channel_unref(conn->channel);
  g_free(conn);
  backend->conns = g_slist_remove(backend->conns, conn);
  return FALSE;
}

static gboolean
c4s_comm_unix_accept_func(G_GNUC_UNUSED GIOChannel* source,
                          GIOCondition condition,
                          gpointer user_data)
{
  C4sCommConn* conn;
  C4sCommBackend* backend;
  int client_sock;

  backend = user_data;
  
  if(condition != G_IO_IN) return FALSE;
  
  client_sock = accept(g_io_channel_unix_get_fd(backend->channel), NULL, NULL);
  if(client_sock == -1) return TRUE;
  
  conn = g_new(C4sCommConn, 1);
  conn->backend = backend;
  conn->channel = c4s_comm_unix_make_io_channel(client_sock);
  conn->size = 0;
  
  backend->conns = g_slist_prepend(backend->conns, conn);
  
  g_io_add_watch(
    conn->channel,
    G_IO_IN,
    c4s_comm_unix_recv_func,
    conn
  );

  return TRUE;
}

static C4sCommBackend*
c4s_comm_unix_init(C4sCommFileFunc file_func,
                   gpointer user_data)
{
  C4sCommBackend* backend;
  backend = g_new(C4sCommBackend, 1);
  
  backend->file_func = file_func;
  backend->user_data = user_data;

  backend->channel = NULL;
  backend->addr.sun_family = AF_UNIX;
  backend->conns = NULL;
  
  return backend;
}

static void
c4s_comm_unix_deinit(C4sCommBackend* backend)
{
  GSList* item;
  C4sCommConn* conn;
  
  for(item = backend->conns; item != NULL; item = item->next)
  {
    conn = item->data;
    g_io_channel_unref(conn->channel);
    g_free(conn);
  }
  
  g_slist_free(backend->conns);

  if(backend->channel != NULL)
  {
    g_io_channel_unref(backend->channel);
    if(backend->host == TRUE)
      unlink(backend->addr.sun_path);
  }
}

static gboolean
c4s_comm_unix_open(C4sCommBackend* backend,
                   G_GNUC_UNUSED GtkWindow* window)
{
  int sock;

  sprintf(
    backend->addr.sun_path,
    "%s/c4s_%d.sock",
    g_get_tmp_dir(),
    getpid()
  );

  backend->host = TRUE;

  sock = socket(PF_UNIX, SOCK_STREAM, 0);
  if(sock == -1) return FALSE;

  if(bind(sock, (struct sockaddr*)&backend->addr, sizeof(backend->addr)) == -1)
    goto error;

  if(listen(sock, 5) == -1)
    goto error;

  backend->channel = c4s_comm_unix_make_io_channel(sock);

  g_io_add_watch(
    backend->channel,
    G_IO_IN,
    c4s_comm_unix_accept_func,
    backend
  );

  return TRUE;

error:
  close(sock);
  return FALSE;
}

static gboolean
c4s_comm_unix_connect(C4sCommBackend* backend)
{
  int sock;
  GPatternSpec* pattern;
  GDir* dir;
  const gchar* file;
  struct stat stat_buf;
  int res;

  pattern = NULL;
  dir = NULL;

  sock = socket(PF_UNIX, SOCK_STREAM, 0);
  if(sock == -1) goto done;
  
  dir = g_dir_open(g_get_tmp_dir(), 0, NULL);
  if(dir == NULL) goto done;

  pattern = g_pattern_spec_new("c4s_*.sock");
  backend->host = FALSE;
  while((file = g_dir_read_name(dir)) != NULL)
  {
    if(g_pattern_match_string(pattern, file) == FALSE)
      continue;

    if(strlen(file) + strlen(g_get_tmp_dir() + 1) >= UNIX_PATH_MAX) continue;
    sprintf(backend->addr.sun_path, "%s/%s", g_get_tmp_dir(), file);

    stat(backend->addr.sun_path, &stat_buf);
    if(stat_buf.st_uid != getuid()) continue;

    /* Not POSIX? */
    /*if(stat_buf.st_mode & S_IFSOCK) == 0) continue;*/
    
    res = connect(
      sock,
      (struct sockaddr*)&backend->addr,
      sizeof(backend->addr)
    );
    
    if(res == -1)
    {
      unlink(backend->addr.sun_path);
    }
    else
    {
      backend->channel = c4s_comm_unix_make_io_channel(sock);
      break;
    }
  }

done:
  if(pattern) g_pattern_spec_free(pattern);
  if(dir) g_dir_close(dir);

  if(backend->channel != NULL) return TRUE;
  if(sock != -1) close(sock);
  return FALSE;
}

static void
c4s_comm_unix_send(C4sCommBackend* backend,
                   int argc,
                   gchar* const* argv)
{
  GString* string;
  gsize bytes_written;
  gsize res;
  GIOStatus status;
  int i;

  if(argc == 0) return;

  /* Make guess about size based on first file */
  string = g_string_sized_new(argc * (strlen(argv[0]) + 1));

  for(i = 0; i < argc; ++ i)
  {
    g_string_append(string, argv[i]);
    g_string_append_c(string, '\n');
  }
  
  bytes_written = 0;
  while(bytes_written < string->len)
  {
    status = g_io_channel_write_chars(
      backend->channel,
      string->str + bytes_written,
      string->len - bytes_written,
      &res,
      NULL
    );

    if(status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN)
      break;

    bytes_written += res;
  }
  
  g_string_free(string, TRUE);
}

const C4sCommBackendFuncs C4S_COMM_BACKEND_UNIX = {
  c4s_comm_unix_init,
  c4s_comm_unix_deinit,
  c4s_comm_unix_open,
  c4s_comm_unix_connect,
  c4s_comm_unix_send
};
